@@ -16,15 +16,26 @@ import os, pickle, re
 from .. import build
 from .. import dependencies
 from .. import mesonlib
+from .. import mlog
 from .. import compilers
 import json
 import subprocess
 from ..mesonlib import MesonException, get_compiler_for_source, classify_unity_sources
+from ..compilers import CompilerArgs
 
-class InstallData():
+class CleanTrees:
+    '''
+    Directories outputted by custom targets that have to be manually cleaned
+    because on Linux `ninja clean` only deletes empty directories.
+    '''
+    def __init__(self, build_dir, trees):
+        self.build_dir = build_dir
+        self.trees = trees
+
+class InstallData:
     def __init__(self, source_dir, build_dir, prefix):
         self.source_dir = source_dir
-        self.build_dir= build_dir
+        self.build_dir = build_dir
         self.prefix = prefix
         self.targets = []
         self.headers = []
@@ -35,7 +46,7 @@ class InstallData():
         self.install_scripts = []
         self.install_subdirs = []
 
-class ExecutableSerialisation():
+class ExecutableSerialisation:
     def __init__(self, name, fname, cmd_args, env, is_cross, exe_wrapper,
                  workdir, extra_paths, capture):
         self.name = name
@@ -50,7 +61,7 @@ class ExecutableSerialisation():
 
 class TestSerialisation:
     def __init__(self, name, suite, fname, is_cross, exe_wrapper, is_parallel, cmd_args, env,
-                 should_fail, valgrind_args, timeout, workdir, extra_paths):
+                 should_fail, timeout, workdir, extra_paths):
         self.name = name
         self.suite = suite
         self.fname = fname
@@ -60,14 +71,13 @@ class TestSerialisation:
         self.cmd_args = cmd_args
         self.env = env
         self.should_fail = should_fail
-        self.valgrind_args = valgrind_args
         self.timeout = timeout
         self.workdir = workdir
         self.extra_paths = extra_paths
 
 # This class contains the basic functionality that is needed by all backends.
 # Feel free to move stuff in and out of it as you see fit.
-class Backend():
+class Backend:
     def __init__(self, build):
         self.build = build
         self.environment = build.environment
@@ -81,7 +91,7 @@ class Backend():
     def get_target_filename(self, t):
         if isinstance(t, build.CustomTarget):
             if len(t.get_outputs()) != 1:
-                mlog.warning('custom_target {!r} has more than one output! ' \
+                mlog.warning('custom_target {!r} has more than one output! '
                              'Using the first one.'.format(t.name))
             filename = t.get_outputs()[0]
         else:
@@ -161,7 +171,6 @@ class Backend():
 
         # For each language, generate a unity source file and return the list
         for comp, srcs in compsrcs.items():
-            lang = comp.get_language()
             with init_language_file(comp.get_default_suffix()) as ofile:
                 for src in srcs:
                     ofile.write('#include<%s>\n' % src)
@@ -169,7 +178,7 @@ class Backend():
         return result
 
     def relpath(self, todir, fromdir):
-        return os.path.relpath(os.path.join('dummyprefixdir', todir),\
+        return os.path.relpath(os.path.join('dummyprefixdir', todir),
                                os.path.join('dummyprefixdir', fromdir))
 
     def flatten_object_list(self, target, proj_dir_to_build_root=''):
@@ -179,6 +188,8 @@ class Backend():
                 o = os.path.join(proj_dir_to_build_root,
                                  self.build_to_src, target.get_subdir(), obj)
                 obj_list.append(o)
+            elif isinstance(obj, mesonlib.File):
+                obj_list.append(obj.rel_to_builddir(self.build_to_src))
             elif isinstance(obj, build.ExtractedObjects):
                 obj_list += self.determine_ext_objs(obj, proj_dir_to_build_root)
             else:
@@ -187,15 +198,20 @@ class Backend():
 
     def serialise_executable(self, exe, cmd_args, workdir, env={},
                              capture=None):
-        import uuid
+        import hashlib
         # Can't just use exe.name here; it will likely be run more than once
         if isinstance(exe, (dependencies.ExternalProgram,
                             build.BuildTarget, build.CustomTarget)):
             basename = exe.name
         else:
             basename = os.path.basename(exe)
-        scratch_file = 'meson_exe_{0}_{1}.dat'.format(basename,
-                                                      str(uuid.uuid4())[:8])
+        # Take a digest of the cmd args, env, workdir, and capture. This avoids
+        # collisions and also makes the name deterministic over regenerations
+        # which avoids a rebuild by Ninja because the cmdline stays the same.
+        data = bytes(str(sorted(env.items())) + str(cmd_args) + str(workdir) + str(capture),
+                     encoding='utf-8')
+        digest = hashlib.sha1(data).hexdigest()
+        scratch_file = 'meson_exe_{0}_{1}.dat'.format(basename, digest)
         exe_data = os.path.join(self.environment.get_scratch_dir(), scratch_file)
         with open(exe_data, 'wb') as f:
             if isinstance(exe, dependencies.ExternalProgram):
@@ -232,41 +248,23 @@ class Backend():
         benchmark_data = os.path.join(self.environment.get_scratch_dir(), 'meson_benchmark_setup.dat')
         with open(benchmark_data, 'wb') as datafile:
             self.write_benchmark_file(datafile)
-        return (test_data, benchmark_data)
+        return test_data, benchmark_data
 
-    def determine_linker(self, target, src):
+    def determine_linker(self, target):
+        '''
+        If we're building a static library, there is only one static linker.
+        Otherwise, we query the target for the dynamic linker.
+        '''
         if isinstance(target, build.StaticLibrary):
             if target.is_cross:
                 return self.build.static_cross_linker
             else:
                 return self.build.static_linker
-        if target.is_cross:
-            compilers = self.build.cross_compilers
-        else:
-            compilers = self.build.compilers
-        if len(compilers) == 1:
-            return compilers[0]
-        # Currently a bit naive. C++ must
-        # be linked with a C++ compiler, but
-        # otherwise we don't care. This will
-        # become trickier if and when Fortran
-        # and the like become supported.
-        cpp = None
-        for c in compilers:
-            if c.get_language() == 'cpp':
-                cpp = c
-                break
-        if cpp is not None:
-            for s in src:
-                if c.can_compile(s):
-                    return cpp
-        for c in compilers:
-            if c.get_language() == 'vala':
-                continue
-            for s in src:
-                if c.can_compile(s):
-                    return c
-        raise AssertionError("BUG: Couldn't determine linker for sources {!r}".format(src))
+        l = target.get_clike_dynamic_linker()
+        if not l:
+            m = "Couldn't determine linker for target {!r}"
+            raise MesonException(m.format(target.name))
+        return l
 
     def object_filename_from_source(self, target, source):
         if isinstance(source, mesonlib.File):
@@ -280,13 +278,9 @@ class Backend():
         result = []
         targetdir = self.get_target_private_dir(extobj.target)
         # With unity builds, there's just one object that contains all the
-        # sources, so if we want all the objects, just return that.
+        # sources, and we only support extracting all the objects in this mode,
+        # so just return that.
         if self.environment.coredata.get_builtin_option('unity'):
-            if not extobj.unity_compatible:
-                # This should never happen
-                msg = 'BUG: Meson must not allow extracting single objects ' \
-                      'in Unity builds'
-                raise AssertionError(msg)
             comp = get_compiler_for_source(extobj.target.compilers.values(),
                                            extobj.srclist[0])
             # The unity object name uses the full absolute path of the source file
@@ -345,32 +339,59 @@ class Backend():
         return extra_args
 
     def generate_basic_compiler_args(self, target, compiler, no_warn_args=False):
-        commands = []
+        # Create an empty commands list, and start adding arguments from
+        # various sources in the order in which they must override each other
+        # starting from hard-coded defaults followed by build options and so on.
+        commands = CompilerArgs(compiler)
+        # First, the trivial ones that are impossible to override.
+        #
+        # Add -nostdinc/-nostdinc++ if needed; can't be overriden
         commands += self.get_cross_stdlib_args(target, compiler)
+        # Add things like /NOLOGO or -pipe; usually can't be overriden
         commands += compiler.get_always_args()
+        # Only add warning-flags by default if the buildtype enables it, and if
+        # we weren't explicitly asked to not emit warnings (for Vala, f.ex)
         if no_warn_args:
             commands += compiler.get_no_warn_args()
-        else:
+        elif self.environment.coredata.get_builtin_option('buildtype') != 'plain':
             commands += compiler.get_warn_args(self.environment.coredata.get_builtin_option('warning_level'))
-        commands += compiler.get_option_compile_args(self.environment.coredata.compiler_options)
-        commands += self.build.get_global_args(compiler)
-        commands += self.build.get_project_args(compiler, target.subproject)
-        commands += self.environment.coredata.external_args[compiler.get_language()]
-        commands += self.escape_extra_args(compiler, target.get_extra_args(compiler.get_language()))
-        commands += compiler.get_buildtype_args(self.environment.coredata.get_builtin_option('buildtype'))
+        # Add -Werror if werror=true is set in the build options set on the
+        # command-line or default_options inside project(). This only sets the
+        # action to be done for warnings if/when they are emitted, so it's ok
+        # to set it after get_no_warn_args() or get_warn_args().
         if self.environment.coredata.get_builtin_option('werror'):
             commands += compiler.get_werror_args()
+        # Add compile args for c_* or cpp_* build options set on the
+        # command-line or default_options inside project().
+        commands += compiler.get_option_compile_args(self.environment.coredata.compiler_options)
+        # Add buildtype args: optimization level, debugging, etc.
+        commands += compiler.get_buildtype_args(self.environment.coredata.get_builtin_option('buildtype'))
+        # Add compile args added using add_project_arguments()
+        commands += self.build.get_project_args(compiler, target.subproject)
+        # Add compile args added using add_global_arguments()
+        # These override per-project arguments
+        commands += self.build.get_global_args(compiler)
+        # Compile args added from the env: CFLAGS/CXXFLAGS, etc. We want these
+        # to override all the defaults, but not the per-target compile args.
+        commands += self.environment.coredata.external_args[compiler.get_language()]
+        # Always set -fPIC for shared libraries
         if isinstance(target, build.SharedLibrary):
             commands += compiler.get_pic_args()
+        # Set -fPIC for static libraries by default unless explicitly disabled
         if isinstance(target, build.StaticLibrary) and target.pic:
             commands += compiler.get_pic_args()
+        # Add compile args needed to find external dependencies
+        # Link args are added while generating the link command
         for dep in target.get_external_deps():
-            # Cflags required by external deps might have UNIX-specific flags,
-            # so filter them out if needed
-            commands += compiler.unix_compile_flags_to_native(dep.get_compile_args())
+            commands += dep.get_compile_args()
+            # Qt needs -fPIC for executables
+            # XXX: We should move to -fPIC for all executables
             if isinstance(target, build.Executable):
-                commands += dep.get_exe_args()
-
+                commands += dep.get_exe_args(compiler)
+            # For 'automagic' deps: Boost and GTest. Also dependency('threads').
+            # pkg-config puts the thread flags itself via `Cflags:`
+            if dep.need_threads():
+                commands += compiler.thread_flags()
         # Fortran requires extra include directives.
         if compiler.language == 'fortran':
             for lt in target.link_targets:
@@ -385,7 +406,7 @@ class Backend():
             if not isinstance(d, (build.StaticLibrary, build.SharedLibrary)):
                 raise RuntimeError('Tried to link with a non-library target "%s".' % d.get_basename())
             if isinstance(compiler, compilers.LLVMDCompiler):
-                args.extend(['-L', self.get_target_filename_for_linking(d)])
+                args += ['-L' + self.get_target_filename_for_linking(d)]
             else:
                 args.append(self.get_target_filename_for_linking(d))
             # If you have executable e that links to shared lib s1 that links to shared library s2
@@ -439,16 +460,23 @@ class Backend():
                 extra_paths = []
             cmd_args = []
             for a in t.cmd_args:
+                if hasattr(a, 'held_object'):
+                    a = a.held_object
                 if isinstance(a, mesonlib.File):
                     a = os.path.join(self.environment.get_build_dir(), a.rel_to_builddir(self.build_to_src))
-                cmd_args.append(a)
+                    cmd_args.append(a)
+                elif isinstance(a, str):
+                    cmd_args.append(a)
+                elif isinstance(a, build.Target):
+                    cmd_args.append(self.get_target_filename(a))
+                else:
+                    raise MesonException('Bad object in test command.')
             ts = TestSerialisation(t.get_name(), t.suite, fname, is_cross, exe_wrapper,
-                                   t.is_parallel, cmd_args, t.env, t.should_fail, t.valgrind_args,
+                                   t.is_parallel, cmd_args, t.env, t.should_fail,
                                    t.timeout, t.workdir, extra_paths)
             arr.append(ts)
         pickle.dump(arr, datafile)
 
-
     def generate_depmf_install(self, d):
         if self.build.dep_manifest_name is None:
             return
@@ -459,12 +487,13 @@ class Backend():
         mfobj['projects'] = self.build.dep_manifest
         with open(ifilename, 'w') as f:
             f.write(json.dumps(mfobj))
-        d.data.append([ifilename, ofilename])
+        # Copy file from, to, and with mode unchanged
+        d.data.append([ifilename, ofilename, None])
 
     def get_regen_filelist(self):
         '''List of all files whose alteration means that the build
         definition needs to be regenerated.'''
-        deps = [os.path.join(self.build_to_src, df) \
+        deps = [os.path.join(self.build_to_src, df)
                 for df in self.interpreter.get_build_def_files()]
         if self.environment.is_cross_build():
             deps.append(os.path.join(self.build_to_src,
@@ -515,6 +544,29 @@ class Backend():
             newargs.append(arg)
         return newargs
 
+    def get_build_by_default_targets(self):
+        result = {}
+        # Get all build and custom targets that must be built by default
+        for name, t in self.build.get_targets().items():
+            if t.build_by_default or t.install or t.build_always:
+                result[name] = t
+        # Get all targets used as test executables and arguments. These must
+        # also be built by default. XXX: Sometime in the future these should be
+        # built only before running tests.
+        for t in self.build.get_tests():
+            exe = t.exe
+            if hasattr(exe, 'held_object'):
+                exe = exe.held_object
+            if isinstance(exe, (build.CustomTarget, build.BuildTarget)):
+                result[exe.get_id()] = exe
+            for arg in t.cmd_args:
+                if hasattr(arg, 'held_object'):
+                    arg = arg.held_object
+                if not isinstance(arg, (build.CustomTarget, build.BuildTarget)):
+                    continue
+                result[arg.get_id()] = arg
+        return result
+
     def get_custom_target_provided_libraries(self, target):
         libs = []
         for t in target.get_generated_sources():
@@ -525,19 +577,13 @@ class Backend():
                     libs.append(os.path.join(self.get_target_dir(t), f))
         return libs
 
-    def eval_custom_target_command(self, target, absolute_paths=False):
-        if not absolute_paths:
-            ofilenames = [os.path.join(self.get_target_dir(target), i) for i in target.output]
-        else:
-            ofilenames = [os.path.join(self.environment.get_build_dir(), self.get_target_dir(target), i) \
-                          for i in target.output]
+    def get_custom_target_sources(self, target):
+        '''
+        Custom target sources can be of various object types; strings, File,
+        BuildTarget, even other CustomTargets.
+        Returns the path to them relative to the build root directory.
+        '''
         srcs = []
-        outdir = self.get_target_dir(target)
-        # Many external programs fail on empty arguments.
-        if outdir == '':
-            outdir = '.'
-        if absolute_paths:
-            outdir = os.path.join(self.environment.get_build_dir(), outdir)
         for i in target.get_sources():
             if hasattr(i, 'held_object'):
                 i = i.held_object
@@ -551,9 +597,25 @@ class Backend():
                 fname = [os.path.join(self.get_target_private_dir(target), p) for p in i.get_outputs()]
             else:
                 fname = [i.rel_to_builddir(self.build_to_src)]
-            if absolute_paths:
-                fname =[os.path.join(self.environment.get_build_dir(), f) for f in fname]
+            if target.absolute_paths:
+                fname = [os.path.join(self.environment.get_build_dir(), f) for f in fname]
             srcs += fname
+        return srcs
+
+    def eval_custom_target_command(self, target, absolute_outputs=False):
+        # We only want the outputs to be absolute when using the VS backend
+        if not absolute_outputs:
+            ofilenames = [os.path.join(self.get_target_dir(target), i) for i in target.output]
+        else:
+            ofilenames = [os.path.join(self.environment.get_build_dir(), self.get_target_dir(target), i)
+                          for i in target.output]
+        srcs = self.get_custom_target_sources(target)
+        outdir = self.get_target_dir(target)
+        # Many external programs fail on empty arguments.
+        if outdir == '':
+            outdir = '.'
+        if target.absolute_paths:
+            outdir = os.path.join(self.environment.get_build_dir(), outdir)
         cmd = []
         for i in target.command:
             if isinstance(i, build.Executable):
@@ -566,9 +628,9 @@ class Backend():
                 i = os.path.join(self.get_target_dir(i), tmp)
             elif isinstance(i, mesonlib.File):
                 i = i.rel_to_builddir(self.build_to_src)
-                if absolute_paths:
+                if target.absolute_paths:
                     i = os.path.join(self.environment.get_build_dir(), i)
-            # FIXME: str types are blindly added and ignore the 'absolute_paths' argument
+            # FIXME: str types are blindly added ignoring 'target.absolute_paths'
             elif not isinstance(i, str):
                 err_msg = 'Argument {0} is of unknown type {1}'
                 raise RuntimeError(err_msg.format(str(i), str(type(i))))
@@ -608,13 +670,13 @@ class Backend():
                 dfilename = os.path.join(outdir, target.depfile)
                 i = i.replace('@DEPFILE@', dfilename)
             elif '@PRIVATE_OUTDIR_' in i:
-                match = re.search('@PRIVATE_OUTDIR_(ABS_)?([^\/\s*]*)@', i)
+                match = re.search('@PRIVATE_OUTDIR_(ABS_)?([^/\s*]*)@', i)
                 if not match:
                     msg = 'Custom target {!r} has an invalid argument {!r}' \
                           ''.format(target.name, i)
                     raise MesonException(msg)
                 source = match.group(0)
-                if match.group(1) is None and not absolute_paths:
+                if match.group(1) is None and not target.absolute_paths:
                     lead_dir = ''
                 else:
                     lead_dir = self.environment.get_build_dir()
@@ -641,36 +703,15 @@ class Backend():
         #
         # https://github.com/mesonbuild/meson/pull/737
         cmd = [i.replace('\\', '/') for i in cmd]
-        return (srcs, ofilenames, cmd)
+        return srcs, ofilenames, cmd
 
     def run_postconf_scripts(self):
-        env = {'MESON_SOURCE_ROOT' : self.environment.get_source_dir(),
-               'MESON_BUILD_ROOT' : self.environment.get_build_dir()
-              }
+        env = {'MESON_SOURCE_ROOT': self.environment.get_source_dir(),
+               'MESON_BUILD_ROOT': self.environment.get_build_dir(),
+               }
         child_env = os.environ.copy()
         child_env.update(env)
 
         for s in self.build.postconf_scripts:
-            cmd = s['exe'].get_command() + s['args']
+            cmd = s['exe'] + s['args']
             subprocess.check_call(cmd, env=child_env)
-
-    # Subprojects of subprojects may cause the same dep args to be used
-    # multiple times. Remove duplicates here. Note that we can't dedup
-    # libraries based on name alone, because "-lfoo -lbar -lfoo" is 
-    # a completely valid (though pathological) sequence and removing the
-    # latter may fail. Usually only applies to static libs, though.
-    def dedup_arguments(self, commands):
-        includes = {}
-        final_commands = []
-        previous = '-fsuch_arguments=woof'
-        for c in commands:
-            if c.startswith(('-I', '-L', '/LIBPATH')):
-                if c in includes:
-                    continue
-                includes[c] = True
-            if previous == c:
-                continue
-            previous = c
-            final_commands.append(c)
-        return final_commands
-                                                  
