@@ -14,7 +14,6 @@
 
 import os, sys
 import pickle
-import re
 import xml.dom.minidom
 import xml.etree.ElementTree as ET
 
@@ -24,6 +23,7 @@ from .. import dependencies
 from .. import mlog
 from .. import compilers
 from ..build import BuildTarget
+from ..compilers import CompilerArgs
 from ..mesonlib import MesonException, File
 from ..environment import Environment
 
@@ -48,7 +48,7 @@ def split_o_flags_args(args):
             o_flags += ['/O' + f for f in flags]
     return o_flags
 
-class RegenInfo():
+class RegenInfo:
     def __init__(self, source_dir, build_dir, depfiles):
         self.source_dir = source_dir
         self.build_dir = build_dir
@@ -122,11 +122,11 @@ class Vs2010Backend(backends.Backend):
                     else:
                         sole_output = ''
                     curfile = infilelist[i]
-                    infilename = os.path.join(self.environment.get_source_dir(), curfile)
+                    infilename = os.path.join(down, curfile.rel_to_builddir(self.build_to_src))
                     outfiles_rel = genlist.get_outputs_for(curfile)
                     outfiles = [os.path.join(target_private_dir, of) for of in outfiles_rel]
                     generator_output_files += outfiles
-                    args = [x.replace("@INPUT@", infilename).replace('@OUTPUT@', sole_output)\
+                    args = [x.replace("@INPUT@", infilename).replace('@OUTPUT@', sole_output)
                             for x in base_args]
                     args = self.replace_outputs(args, target_private_dir, outfiles_rel)
                     args = [x.replace("@SOURCE_DIR@", self.environment.get_source_dir()).replace("@BUILD_DIR@", target_private_dir)
@@ -192,45 +192,55 @@ class Vs2010Backend(backends.Backend):
         result = {}
         for o in obj_list:
             if isinstance(o, build.ExtractedObjects):
-                result[o.target.get_id()] = True
-        return result.keys()
+                result[o.target.get_id()] = o.target
+        return result.items()
 
-    def determine_deps(self, p):
+    def get_target_deps(self, t, recursive=False):
         all_deps = {}
-        target = self.build.targets[p[0]]
-        if isinstance(target, build.CustomTarget):
-            for d in target.get_target_dependencies():
-                all_deps[d.get_id()] = True
-            return all_deps
-        if isinstance(target, build.RunTarget):
-            for d in [target.command] + target.args:
-                if isinstance(d, build.BuildTarget):
-                    all_deps[d.get_id()] = True
-                return all_deps
-        for ldep in target.link_targets:
-            all_deps[ldep.get_id()] = True
-        for objdep in self.get_obj_target_deps(target.objects):
-            all_deps[objdep] = True
-        for gendep in target.generated:
-            if isinstance(gendep, build.CustomTarget):
-                all_deps[gendep.get_id()] = True
+        for target in t.values():
+            if isinstance(target, build.CustomTarget):
+                for d in target.get_target_dependencies():
+                    all_deps[d.get_id()] = d
+            elif isinstance(target, build.RunTarget):
+                for d in [target.command] + target.args:
+                    if isinstance(d, (build.BuildTarget, build.CustomTarget)):
+                        all_deps[d.get_id()] = d
+            # BuildTarget
             else:
-                gen_exe = gendep.generator.get_exe()
-                if isinstance(gen_exe, build.Executable):
-                    all_deps[gen_exe.get_id()] = True
-        return all_deps
+                for ldep in target.link_targets:
+                    all_deps[ldep.get_id()] = ldep
+                for obj_id, objdep in self.get_obj_target_deps(target.objects):
+                    all_deps[obj_id] = objdep
+                for gendep in target.get_generated_sources():
+                    if isinstance(gendep, build.CustomTarget):
+                        all_deps[gendep.get_id()] = gendep
+                    else:
+                        gen_exe = gendep.generator.get_exe()
+                        if isinstance(gen_exe, build.Executable):
+                            all_deps[gen_exe.get_id()] = gen_exe
+        if not t or not recursive:
+            return all_deps
+        ret = self.get_target_deps(all_deps, recursive)
+        ret.update(all_deps)
+        return ret
 
     def generate_solution(self, sln_filename, projlist):
+        default_projlist = self.get_build_by_default_targets()
         with open(sln_filename, 'w') as ofile:
             ofile.write('Microsoft Visual Studio Solution File, Format '
                         'Version 11.00\n')
             ofile.write('# Visual Studio ' + self.vs_version + '\n')
-            prj_templ = prj_line = 'Project("{%s}") = "%s", "%s", "{%s}"\n'
+            prj_templ = 'Project("{%s}") = "%s", "%s", "{%s}"\n'
             for p in projlist:
                 prj_line = prj_templ % (self.environment.coredata.guid,
                                         p[0], p[1], p[2])
                 ofile.write(prj_line)
-                all_deps = self.determine_deps(p)
+                target = self.build.targets[p[0]]
+                t = {target.get_id(): target}
+                # Get direct deps
+                all_deps = self.get_target_deps(t)
+                # Get recursive deps
+                recursive_deps = self.get_target_deps(t, recursive=True)
                 ofile.write('\tProjectSection(ProjectDependencies) = '
                             'postProject\n')
                 regen_guid = self.environment.coredata.regen_guid
@@ -240,6 +250,9 @@ class Vs2010Backend(backends.Backend):
                     ofile.write('\t\t{%s} = {%s}\n' % (guid, guid))
                 ofile.write('EndProjectSection\n')
                 ofile.write('EndProject\n')
+                for dep, target in recursive_deps.items():
+                    if p[0] in default_projlist:
+                        default_projlist[dep] = target
             test_line = prj_templ % (self.environment.coredata.guid,
                                      'RUN_TESTS', 'RUN_TESTS.vcxproj',
                                      self.environment.coredata.test_guid)
@@ -259,17 +272,21 @@ class Vs2010Backend(backends.Backend):
             ofile.write('\tEndGlobalSection\n')
             ofile.write('\tGlobalSection(ProjectConfigurationPlatforms) = '
                         'postSolution\n')
-            ofile.write('\t\t{%s}.%s|%s.ActiveCfg = %s|%s\n' % 
+            ofile.write('\t\t{%s}.%s|%s.ActiveCfg = %s|%s\n' %
                         (self.environment.coredata.regen_guid, self.buildtype,
                          self.platform, self.buildtype, self.platform))
             ofile.write('\t\t{%s}.%s|%s.Build.0 = %s|%s\n' %
                         (self.environment.coredata.regen_guid, self.buildtype,
                          self.platform, self.buildtype, self.platform))
+            # Create the solution configuration
             for p in projlist:
+                # Add to the list of projects in this solution
                 ofile.write('\t\t{%s}.%s|%s.ActiveCfg = %s|%s\n' %
                             (p[2], self.buildtype, self.platform,
                              self.buildtype, self.platform))
-                if not isinstance(self.build.targets[p[0]], build.RunTarget):
+                if p[0] in default_projlist and \
+                   not isinstance(self.build.targets[p[0]], build.RunTarget):
+                    # Add to the list of projects to be built
                     ofile.write('\t\t{%s}.%s|%s.Build.0 = %s|%s\n' %
                                 (p[2], self.buildtype, self.platform,
                                  self.buildtype, self.platform))
@@ -314,28 +331,28 @@ class Vs2010Backend(backends.Backend):
             else:
                 # Everything that is not an object or source file is considered a header.
                 headers.append(i)
-        return (sources, headers, objects, languages)
+        return sources, headers, objects, languages
 
     def target_to_build_root(self, target):
         if target.subdir == '':
             return ''
 
         directories = os.path.normpath(target.subdir).split(os.sep)
-        return os.sep.join(['..']*len(directories))
+        return os.sep.join(['..'] * len(directories))
 
     def special_quote(self, arr):
         return ['&quot;%s&quot;' % i for i in arr]
 
     def create_basic_crap(self, target):
         project_name = target.name
-        root = ET.Element('Project', {'DefaultTargets' : "Build",
-                                      'ToolsVersion' : '4.0',
-                                      'xmlns' : 'http://schemas.microsoft.com/developer/msbuild/2003'})
-        confitems = ET.SubElement(root, 'ItemGroup', {'Label' : 'ProjectConfigurations'})
+        root = ET.Element('Project', {'DefaultTargets': "Build",
+                                      'ToolsVersion': '4.0',
+                                      'xmlns': 'http://schemas.microsoft.com/developer/msbuild/2003'})
+        confitems = ET.SubElement(root, 'ItemGroup', {'Label': 'ProjectConfigurations'})
         prjconf = ET.SubElement(confitems, 'ProjectConfiguration',
-                                {'Include' : self.buildtype + '|' + self.platform})
+                                {'Include': self.buildtype + '|' + self.platform})
         p = ET.SubElement(prjconf, 'Configuration')
-        p.text= self.buildtype
+        p.text = self.buildtype
         pl = ET.SubElement(prjconf, 'Platform')
         pl.text = self.platform
         globalgroup = ET.SubElement(root, 'PropertyGroup', Label='Globals')
@@ -344,8 +361,8 @@ class Vs2010Backend(backends.Backend):
         kw = ET.SubElement(globalgroup, 'Keyword')
         kw.text = self.platform + 'Proj'
         p = ET.SubElement(globalgroup, 'Platform')
-        p.text= self.platform
-        pname= ET.SubElement(globalgroup, 'ProjectName')
+        p.text = self.platform
+        pname = ET.SubElement(globalgroup, 'ProjectName')
         pname.text = project_name
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.Default.props')
         type_config = ET.SubElement(root, 'PropertyGroup', Label='Configuration')
@@ -381,7 +398,7 @@ class Vs2010Backend(backends.Backend):
                 cmd += i.fullpath
             else:
                 cmd.append(i)
-        cmd_templ = '''"%s" '''*len(cmd)
+        cmd_templ = '''"%s" ''' * len(cmd)
         ET.SubElement(customstep, 'Command').text = cmd_templ % tuple(cmd)
         ET.SubElement(customstep, 'Message').text = 'Running custom command.'
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.targets')
@@ -396,7 +413,7 @@ class Vs2010Backend(backends.Backend):
         # from the target dir, not the build root.
         target.absolute_paths = True
         (srcs, ofilenames, cmd) = self.eval_custom_target_command(target, True)
-        cmd_templ = '''"%s" '''*len(cmd)
+        cmd_templ = '''"%s" ''' * len(cmd)
         ET.SubElement(customstep, 'Command').text = cmd_templ % tuple(cmd)
         ET.SubElement(customstep, 'Outputs').text = ';'.join(ofilenames)
         ET.SubElement(customstep, 'Inputs').text = ';'.join(srcs)
@@ -427,24 +444,25 @@ class Vs2010Backend(backends.Backend):
         pch_out.text = '$(IntDir)$(TargetName)-%s.pch' % lang
 
     def add_additional_options(self, lang, parent_node, file_args):
-        if len(file_args[lang]) == 0:
-            # We only need per file options if they were not set per project.
-            return
-        args = file_args[lang] + ['%(AdditionalOptions)']
+        args = []
+        for arg in file_args[lang].to_native():
+            if arg == '%(AdditionalOptions)':
+                args.append(arg)
+            else:
+                args.append(self.escape_additional_option(arg))
         ET.SubElement(parent_node, "AdditionalOptions").text = ' '.join(args)
 
     def add_preprocessor_defines(self, lang, parent_node, file_defines):
-        if len(file_defines[lang]) == 0:
-            # We only need per file options if they were not set per project.
-            return
-        defines = file_defines[lang] + ['%(PreprocessorDefinitions)']
+        defines = []
+        for define in file_defines[lang]:
+            if define == '%(PreprocessorDefinitions)':
+                defines.append(define)
+            else:
+                defines.append(self.escape_preprocessor_define(define))
         ET.SubElement(parent_node, "PreprocessorDefinitions").text = ';'.join(defines)
 
     def add_include_dirs(self, lang, parent_node, file_inc_dirs):
-        if len(file_inc_dirs[lang]) == 0:
-            # We only need per file options if they were not set per project.
-            return
-        dirs = file_inc_dirs[lang] + ['%(AdditionalIncludeDirectories)']
+        dirs = file_inc_dirs[lang]
         ET.SubElement(parent_node, "AdditionalIncludeDirectories").text = ';'.join(dirs)
 
     @staticmethod
@@ -463,18 +481,18 @@ class Vs2010Backend(backends.Backend):
     def escape_preprocessor_define(define):
         # See: https://msdn.microsoft.com/en-us/library/bb383819.aspx
         table = str.maketrans({'%': '%25', '$': '%24', '@': '%40',
-            "'": '%27', ';': '%3B', '?': '%3F', '*': '%2A',
-            # We need to escape backslash because it'll be un-escaped by
-            # Windows during process creation when it parses the arguments
-            # Basically, this converts `\` to `\\`.
-            '\\': '\\\\'})
+                               "'": '%27', ';': '%3B', '?': '%3F', '*': '%2A',
+                               # We need to escape backslash because it'll be un-escaped by
+                               # Windows during process creation when it parses the arguments
+                               # Basically, this converts `\` to `\\`.
+                               '\\': '\\\\'})
         return define.translate(table)
 
     @staticmethod
     def escape_additional_option(option):
         # See: https://msdn.microsoft.com/en-us/library/bb383819.aspx
         table = str.maketrans({'%': '%25', '$': '%24', '@': '%40',
-            "'": '%27', ';': '%3B', '?': '%3F', '*': '%2A', ' ': '%20',})
+                               "'": '%27', ';': '%3B', '?': '%3F', '*': '%2A', ' ': '%20'})
         option = option.translate(table)
         # Since we're surrounding the option with ", if it ends in \ that will
         # escape the " when the process arguments are parsed and the starting
@@ -513,7 +531,7 @@ class Vs2010Backend(backends.Backend):
                     libs.append(arg)
             else:
                 other.append(arg)
-        return (lpaths, libs, other)
+        return lpaths, libs, other
 
     def _get_cl_compiler(self, target):
         for lang, c in target.compilers.items():
@@ -559,14 +577,14 @@ class Vs2010Backend(backends.Backend):
         buildtype_link_args = compiler.get_buildtype_linker_args(self.buildtype)
         project_name = target.name
         target_name = target.name
-        root = ET.Element('Project', {'DefaultTargets' : "Build",
-                                      'ToolsVersion' : '4.0',
-                                      'xmlns' : 'http://schemas.microsoft.com/developer/msbuild/2003'})
-        confitems = ET.SubElement(root, 'ItemGroup', {'Label' : 'ProjectConfigurations'})
+        root = ET.Element('Project', {'DefaultTargets': "Build",
+                                      'ToolsVersion': '4.0',
+                                      'xmlns': 'http://schemas.microsoft.com/developer/msbuild/2003'})
+        confitems = ET.SubElement(root, 'ItemGroup', {'Label': 'ProjectConfigurations'})
         prjconf = ET.SubElement(confitems, 'ProjectConfiguration',
-                                {'Include' : self.buildtype + '|' + self.platform})
+                                {'Include': self.buildtype + '|' + self.platform})
         p = ET.SubElement(prjconf, 'Configuration')
-        p.text= self.buildtype
+        p.text = self.buildtype
         pl = ET.SubElement(prjconf, 'Platform')
         pl.text = self.platform
         # Globals
@@ -578,8 +596,8 @@ class Vs2010Backend(backends.Backend):
         ns = ET.SubElement(globalgroup, 'RootNamespace')
         ns.text = target_name
         p = ET.SubElement(globalgroup, 'Platform')
-        p.text= self.platform
-        pname= ET.SubElement(globalgroup, 'ProjectName')
+        p.text = self.platform
+        pname = ET.SubElement(globalgroup, 'ProjectName')
         pname.text = project_name
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.Default.props')
         # Start configuration
@@ -669,91 +687,132 @@ class Vs2010Backend(backends.Backend):
         # Arguments, include dirs, defines for all files in the current target
         target_args = []
         target_defines = []
-        target_inc_dirs = ['.', self.relpath(self.get_target_private_dir(target),
-                                             self.get_target_dir(target)),
-                           proj_to_src_dir] + generated_files_include_dirs
+        target_inc_dirs = []
         # Arguments, include dirs, defines passed to individual files in
         # a target; perhaps because the args are language-specific
-        file_args = dict((lang, []) for lang in target.compilers)
+        #
+        # file_args is also later split out into defines and include_dirs in
+        # case someone passed those in there
+        file_args = dict((lang, CompilerArgs(comp)) for lang, comp in target.compilers.items())
         file_defines = dict((lang, []) for lang in target.compilers)
         file_inc_dirs = dict((lang, []) for lang in target.compilers)
-        for l, args in self.environment.coredata.external_args.items():
+        # The order in which these compile args are added must match
+        # generate_single_compile() and generate_basic_compiler_args()
+        for l, comp in target.compilers.items():
             if l in file_args:
-                file_args[l] += args
-        for l, args in self.build.global_args.items():
-            if l in file_args:
-                file_args[l] += args
+                file_args[l] += compilers.get_base_compile_args(self.environment.coredata.base_options, comp)
+                file_args[l] += comp.get_option_compile_args(self.environment.coredata.compiler_options)
+        # Add compile args added using add_project_arguments()
         for l, args in self.build.projects_args.get(target.subproject, {}).items():
             if l in file_args:
                 file_args[l] += args
+        # Add compile args added using add_global_arguments()
+        # These override per-project arguments
+        for l, args in self.build.global_args.items():
+            if l in file_args:
+                file_args[l] += args
+        # Compile args added from the env: CFLAGS/CXXFLAGS, etc. We want these
+        # to override all the defaults, but not the per-target compile args.
+        for l, args in self.environment.coredata.external_args.items():
+            if l in file_args:
+                file_args[l] += args
+        for args in file_args.values():
+            # This is where Visual Studio will insert target_args, target_defines,
+            # etc, which are added later from external deps (see below).
+            args += ['%(AdditionalOptions)', '%(PreprocessorDefinitions)', '%(AdditionalIncludeDirectories)']
+            # Add include dirs from the `include_directories:` kwarg on the target
+            # and from `include_directories:` of internal deps of the target.
+            #
+            # Target include dirs should override internal deps include dirs.
+            #
+            # Include dirs from internal deps should override include dirs from
+            # external deps.
+            # These are per-target, but we still add them as per-file because we
+            # need them to be looked in first.
+            for d in target.get_include_dirs():
+                for i in d.get_incdirs():
+                    curdir = os.path.join(d.get_curdir(), i)
+                    args.append('-I' + self.relpath(curdir, target.subdir)) # build dir
+                    args.append('-I' + os.path.join(proj_to_src_root, curdir)) # src dir
+                for i in d.get_extra_build_dirs():
+                    curdir = os.path.join(d.get_curdir(), i)
+                    args.append('-I' + self.relpath(curdir, target.subdir))  # build dir
+        # Add per-target compile args, f.ex, `c_args : ['/DFOO']`. We set these
+        # near the end since these are supposed to override everything else.
         for l, args in target.extra_args.items():
             if l in file_args:
-                file_args[l] += compiler.unix_compile_flags_to_native(args)
-        for l, comp in target.compilers.items():
-            if l in file_args:
-                file_args[l] += comp.get_option_compile_args(self.environment.coredata.compiler_options)
+                file_args[l] += args
+        # The highest priority includes. In order of directory search:
+        # target private dir, target build dir, generated sources include dirs,
+        # target source dir
+        for args in file_args.values():
+            t_inc_dirs = ['.', self.relpath(self.get_target_private_dir(target),
+                                            self.get_target_dir(target))]
+            t_inc_dirs += generated_files_include_dirs + [proj_to_src_dir]
+            args += ['-I' + arg for arg in t_inc_dirs]
+
+        # Split preprocessor defines and include directories out of the list of
+        # all extra arguments. The rest go into %(AdditionalOptions).
+        for l, args in file_args.items():
+            for arg in args[:]:
+                if arg.startswith(('-D', '/D')) or arg == '%(PreprocessorDefinitions)':
+                    file_args[l].remove(arg)
+                    # Don't escape the marker
+                    if arg == '%(PreprocessorDefinitions)':
+                        define = arg
+                    else:
+                        define = arg[2:]
+                    # De-dup
+                    if define in file_defines[l]:
+                        file_defines[l].remove(define)
+                    file_defines[l].append(define)
+                elif arg.startswith(('-I', '/I')) or arg == '%(AdditionalIncludeDirectories)':
+                    file_args[l].remove(arg)
+                    # Don't escape the marker
+                    if arg == '%(AdditionalIncludeDirectories)':
+                        inc_dir = arg
+                    else:
+                        inc_dir = arg[2:]
+                    # De-dup
+                    if inc_dir not in file_inc_dirs[l]:
+                        file_inc_dirs[l].append(inc_dir)
+
+        # Split compile args needed to find external dependencies
+        # Link args are added while generating the link command
         for d in target.get_external_deps():
             # Cflags required by external deps might have UNIX-specific flags,
             # so filter them out if needed
-            d_compile_args = compiler.unix_compile_flags_to_native(d.get_compile_args())
+            d_compile_args = compiler.unix_args_to_native(d.get_compile_args())
             for arg in d_compile_args:
                 if arg.startswith(('-D', '/D')):
                     define = arg[2:]
                     # De-dup
-                    if define not in target_defines:
-                        target_defines.append(define)
+                    if define in target_defines:
+                        target_defines.remove(define)
+                    target_defines.append(define)
                 elif arg.startswith(('-I', '/I')):
                     inc_dir = arg[2:]
                     # De-dup
                     if inc_dir not in target_inc_dirs:
                         target_inc_dirs.append(inc_dir)
                 else:
-                    # De-dup
-                    if arg not in target_args:
-                        target_args.append(arg)
-
-        # Split preprocessor defines and include directories out of the list of
-        # all extra arguments. The rest go into %(AdditionalOptions).
-        for l, args in file_args.items():
-            file_args[l] = []
-            for arg in args:
-                if arg.startswith(('-D', '/D')):
-                    define = self.escape_preprocessor_define(arg[2:])
-                    # De-dup
-                    if define not in file_defines[l]:
-                        file_defines[l].append(define)
-                elif arg.startswith(('-I', '/I')):
-                    inc_dir = arg[2:]
-                    # De-dup
-                    if inc_dir not in file_inc_dirs[l]:
-                        file_inc_dirs[l].append(inc_dir)
-                else:
-                    file_args[l].append(self.escape_additional_option(arg))
+                    target_args.append(arg)
 
         languages += gen_langs
         if len(target_args) > 0:
             target_args.append('%(AdditionalOptions)')
             ET.SubElement(clconf, "AdditionalOptions").text = ' '.join(target_args)
-        additional_options_set = True
-
-        for d in target.include_dirs:
-            for i in d.incdirs:
-                curdir = os.path.join(d.curdir, i)
-                target_inc_dirs.append(self.relpath(curdir, target.subdir)) # build dir
-                target_inc_dirs.append(os.path.join(proj_to_src_root, curdir)) # src dir
-            for i in d.get_extra_build_dirs():
-                curdir = os.path.join(d.curdir, i)
-                target_inc_dirs.append(self.relpath(curdir, target.subdir))  # build dir
 
         target_inc_dirs.append('%(AdditionalIncludeDirectories)')
         ET.SubElement(clconf, 'AdditionalIncludeDirectories').text = ';'.join(target_inc_dirs)
         target_defines.append('%(PreprocessorDefinitions)')
         ET.SubElement(clconf, 'PreprocessorDefinitions').text = ';'.join(target_defines)
-        rebuild = ET.SubElement(clconf, 'MinimalRebuild')
-        rebuild.text = 'true'
-        funclink = ET.SubElement(clconf, 'FunctionLevelLinking')
-        funclink.text = 'true'
+        ET.SubElement(clconf, 'MinimalRebuild').text = 'true'
+        ET.SubElement(clconf, 'FunctionLevelLinking').text = 'true'
         pch_node = ET.SubElement(clconf, 'PrecompiledHeader')
+        if self.environment.coredata.get_builtin_option('werror'):
+            ET.SubElement(clconf, 'TreatWarningAsError').text = 'true'
+        # Note: SuppressStartupBanner is /NOLOGO and is 'true' by default
         pch_sources = {}
         for lang in ['c', 'cpp']:
             pch = target.get_pch(lang)
@@ -775,19 +834,27 @@ class Vs2010Backend(backends.Backend):
 
         resourcecompile = ET.SubElement(compiles, 'ResourceCompile')
         ET.SubElement(resourcecompile, 'PreprocessorDefinitions')
+
+        # Linker options
         link = ET.SubElement(compiles, 'Link')
-        # Put all language args here, too.
-        extra_link_args = compiler.get_option_link_args(self.environment.coredata.compiler_options)
+        extra_link_args = CompilerArgs(compiler)
         # FIXME: Can these buildtype linker args be added as tags in the
         # vcxproj file (similar to buildtype compiler args) instead of in
         # AdditionalOptions?
         extra_link_args += compiler.get_buildtype_linker_args(self.buildtype)
-        for l in self.environment.coredata.external_link_args.values():
-            extra_link_args += l
         if not isinstance(target, build.StaticLibrary):
-            extra_link_args += target.link_args
             if isinstance(target, build.SharedModule):
                 extra_link_args += compiler.get_std_shared_module_link_args()
+            # Add link args added using add_project_link_arguments()
+            extra_link_args += self.build.get_project_link_args(compiler, target.subproject)
+            # Add link args added using add_global_link_arguments()
+            # These override per-project link arguments
+            extra_link_args += self.build.get_global_link_args(compiler)
+            # Link args added from the env: LDFLAGS. We want these to
+            # override all the defaults but not the per-target link args.
+            extra_link_args += self.environment.coredata.external_link_args[compiler.get_language()]
+            # Only non-static built targets need link args and link dependencies
+            extra_link_args += target.link_args
             # External deps must be last because target link libraries may depend on them.
             for dep in target.get_external_deps():
                 extra_link_args += dep.get_link_args()
@@ -795,8 +862,13 @@ class Vs2010Backend(backends.Backend):
                 if isinstance(d, build.StaticLibrary):
                     for dep in d.get_external_deps():
                         extra_link_args += dep.get_link_args()
-        extra_link_args = compiler.unix_link_flags_to_native(extra_link_args)
-        (additional_libpaths, additional_links, extra_link_args) = self.split_link_args(extra_link_args)
+        # Add link args for c_* or cpp_* build options. Currently this only
+        # adds c_winlibs and cpp_winlibs when building for Windows. This needs
+        # to be after all internal and external libraries so that unresolved
+        # symbols from those can be found here. This is needed when the
+        # *_winlibs that we want to link to are static mingw64 libraries.
+        extra_link_args += compiler.get_option_link_args(self.environment.coredata.compiler_options)
+        (additional_libpaths, additional_links, extra_link_args) = self.split_link_args(extra_link_args.to_native())
         if len(extra_link_args) > 0:
             extra_link_args.append('%(AdditionalOptions)')
             ET.SubElement(link, "AdditionalOptions").text = ' '.join(extra_link_args)
@@ -884,7 +956,6 @@ class Vs2010Backend(backends.Backend):
                 header, impl, suffix = pch_sources[lang]
                 relpath = os.path.join(proj_to_src_dir, impl)
                 inc_cl = ET.SubElement(inc_src, 'CLCompile', Include=relpath)
-                lang = Vs2010Backend.lang_from_source_file(s)
                 pch = ET.SubElement(inc_cl, 'PrecompiledHeader')
                 pch.text = 'Create'
                 pch_out = ET.SubElement(inc_cl, 'PrecompiledHeaderOutputFile')
@@ -927,13 +998,13 @@ class Vs2010Backend(backends.Backend):
 
     def gen_regenproj(self, project_name, ofname):
         root = ET.Element('Project', {'DefaultTargets': 'Build',
-                                      'ToolsVersion' : '4.0',
-                                      'xmlns' : 'http://schemas.microsoft.com/developer/msbuild/2003'})
-        confitems = ET.SubElement(root, 'ItemGroup', {'Label' : 'ProjectConfigurations'})
-        prjconf = ET.SubElement(confitems, 'ProjectConfiguration', 
-                                {'Include' : self.buildtype + '|' + self.platform})
+                                      'ToolsVersion': '4.0',
+                                      'xmlns': 'http://schemas.microsoft.com/developer/msbuild/2003'})
+        confitems = ET.SubElement(root, 'ItemGroup', {'Label': 'ProjectConfigurations'})
+        prjconf = ET.SubElement(confitems, 'ProjectConfiguration',
+                                {'Include': self.buildtype + '|' + self.platform})
         p = ET.SubElement(prjconf, 'Configuration')
-        p.text= self.buildtype
+        p.text = self.buildtype
         pl = ET.SubElement(prjconf, 'Platform')
         pl.text = self.platform
         globalgroup = ET.SubElement(root, 'PropertyGroup', Label='Globals')
@@ -943,7 +1014,7 @@ class Vs2010Backend(backends.Backend):
         kw.text = self.platform + 'Proj'
         p = ET.SubElement(globalgroup, 'Platform')
         p.text = self.platform
-        pname= ET.SubElement(globalgroup, 'ProjectName')
+        pname = ET.SubElement(globalgroup, 'ProjectName')
         pname.text = project_name
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.Default.props')
         type_config = ET.SubElement(root, 'PropertyGroup', Label='Configuration')
@@ -1005,14 +1076,14 @@ if %%errorlevel%% neq 0 goto :VCEnd'''
 
     def gen_testproj(self, target_name, ofname):
         project_name = target_name
-        root = ET.Element('Project', {'DefaultTargets' : "Build",
-                                      'ToolsVersion' : '4.0',
-                                      'xmlns' : 'http://schemas.microsoft.com/developer/msbuild/2003'})
-        confitems = ET.SubElement(root, 'ItemGroup', {'Label' : 'ProjectConfigurations'})
+        root = ET.Element('Project', {'DefaultTargets': "Build",
+                                      'ToolsVersion': '4.0',
+                                      'xmlns': 'http://schemas.microsoft.com/developer/msbuild/2003'})
+        confitems = ET.SubElement(root, 'ItemGroup', {'Label': 'ProjectConfigurations'})
         prjconf = ET.SubElement(confitems, 'ProjectConfiguration',
-                                {'Include' : self.buildtype + '|' + self.platform})
+                                {'Include': self.buildtype + '|' + self.platform})
         p = ET.SubElement(prjconf, 'Configuration')
-        p.text= self.buildtype
+        p.text = self.buildtype
         pl = ET.SubElement(prjconf, 'Platform')
         pl.text = self.platform
         globalgroup = ET.SubElement(root, 'PropertyGroup', Label='Globals')
@@ -1021,8 +1092,8 @@ if %%errorlevel%% neq 0 goto :VCEnd'''
         kw = ET.SubElement(globalgroup, 'Keyword')
         kw.text = self.platform + 'Proj'
         p = ET.SubElement(globalgroup, 'Platform')
-        p.text= self.platform
-        pname= ET.SubElement(globalgroup, 'ProjectName')
+        p.text = self.platform
+        pname = ET.SubElement(globalgroup, 'ProjectName')
         pname.text = project_name
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.Default.props')
         type_config = ET.SubElement(root, 'PropertyGroup', Label='Configuration')
@@ -1052,12 +1123,19 @@ if %%errorlevel%% neq 0 goto :VCEnd'''
         ET.SubElement(midl, 'ProxyFileName').text = '%(Filename)_p.c'
         postbuild = ET.SubElement(action, 'PostBuildEvent')
         ET.SubElement(postbuild, 'Message')
+        # FIXME: No benchmarks?
+        meson_py = self.environment.get_build_command()
+        (base, ext) = os.path.splitext(meson_py)
+        mesontest_py = base + 'test' + ext
         test_command = [sys.executable,
-                        self.environment.get_build_command(),
-                        '--internal',
-                        'test']
+                        mesontest_py,
+                        '--no-rebuild']
+        if not self.environment.coredata.get_builtin_option('stdsplit'):
+            test_command += ['--no-stdsplit']
+        if self.environment.coredata.get_builtin_option('errorlogs'):
+            test_command += ['--print-errorlogs']
         cmd_templ = '''setlocal
-"%s" "%s"
+"%s"
 if %%errorlevel%% neq 0 goto :cmEnd
 :cmEnd
 endlocal & call :cmErrorLevel %%errorlevel%% & goto :cmDone
@@ -1065,12 +1143,12 @@ endlocal & call :cmErrorLevel %%errorlevel%% & goto :cmDone
 exit /b %%1
 :cmDone
 if %%errorlevel%% neq 0 goto :VCEnd'''
-        test_data = self.serialise_tests()[0]
+        self.serialise_tests()
         ET.SubElement(postbuild, 'Command').text =\
-            cmd_templ % ('" "'.join(test_command), test_data)
+            cmd_templ % ('" "'.join(test_command))
         ET.SubElement(root, 'Import', Project='$(VCTargetsPath)\Microsoft.Cpp.targets')
         tree = ET.ElementTree(root)
         tree.write(ofname, encoding='utf-8', xml_declaration=True)
         # ElementTree can not do prettyprinting so do it manually
-        #doc = xml.dom.minidom.parse(ofname)
-        #open(ofname, 'w').write(doc.toprettyxml())
+        # doc = xml.dom.minidom.parse(ofname)
+        # open(ofname, 'w').write(doc.toprettyxml())
