@@ -1,4 +1,4 @@
-# Copyright 2012-2014 The Meson development team
+# Copyright 2012-2017 The Meson development team
 
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -12,7 +12,6 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from . import coredata
 from . import environment
 from . import dependencies
 from . import mlog
@@ -21,30 +20,32 @@ from .mesonlib import File, flatten, MesonException, stringlistify, classify_uni
 from .environment import for_windows, for_darwin
 from .compilers import is_object, clike_langs, lang_suffixes
 
-known_basic_kwargs = {'install' : True,
-                      'c_pch' : True,
-                      'cpp_pch' : True,
-                      'c_args' : True,
-                      'cpp_args' : True,
-                      'cs_args' : True,
-                      'vala_args' : True,
-                      'd_args' : True,
-                      'java_args' : True,
-                      'link_args' : True,
+known_basic_kwargs = {'install': True,
+                      'c_pch': True,
+                      'cpp_pch': True,
+                      'c_args': True,
+                      'cpp_args': True,
+                      'cs_args': True,
+                      'vala_args': True,
+                      'fortran_args': True,
+                      'd_args': True,
+                      'java_args': True,
+                      'link_args': True,
                       'link_depends': True,
-                      'link_with' : True,
+                      'link_with': True,
                       'include_directories': True,
-                      'dependencies' : True,
-                      'install_dir' : True,
-                      'main_class' : True,
-                      'gui_app' : True,
-                      'extra_files' : True,
-                      'install_rpath' : True,
-                      'resources' : True,
-                      'sources' : True,
-                      'objects' : True,
-                      'native' : True,
-                     }
+                      'dependencies': True,
+                      'install_dir': True,
+                      'main_class': True,
+                      'gui_app': True,
+                      'extra_files': True,
+                      'install_rpath': True,
+                      'resources': True,
+                      'sources': True,
+                      'objects': True,
+                      'native': True,
+                      'build_by_default': True,
+                      }
 
 # These contain kwargs supported by both static and shared libraries. These are
 # combined here because a library() call might be shared_library() or
@@ -52,16 +53,16 @@ known_basic_kwargs = {'install' : True,
 # FIXME: Find a way to pass that info down here so we can have proper target
 # kwargs checking when specifically using shared_library() or static_library().
 known_lib_kwargs = known_basic_kwargs.copy()
-known_lib_kwargs.update({'version' : True, # Only for shared libs
-                         'soversion' : True, # Only for shared libs
-                         'name_prefix' : True,
-                         'name_suffix' : True,
-                         'vs_module_defs' : True, # Only for shared libs
+known_lib_kwargs.update({'version': True, # Only for shared libs
+                         'soversion': True, # Only for shared libs
+                         'name_prefix': True,
+                         'name_suffix': True,
+                         'vs_module_defs': True, # Only for shared libs
                          'vala_header': True,
                          'vala_vapi': True,
-                         'vala_gir' : True,
-                         'pic' : True, # Only for static libs
-                        })
+                         'vala_gir': True,
+                         'pic': True, # Only for static libs
+                         })
 
 
 class InvalidArguments(MesonException):
@@ -98,6 +99,7 @@ class Build:
         self.dep_manifest_name = None
         self.dep_manifest = {}
         self.cross_stdlibs = {}
+        self.test_setups = {}
 
     def add_compiler(self, compiler):
         if self.static_linker is None and compiler.needs_static_linker():
@@ -156,7 +158,7 @@ class Build:
 
         return link_args.get(compiler.get_language(), [])
 
-class IncludeDirs():
+class IncludeDirs:
     def __init__(self, curdir, dirs, is_system, extra_build_dirs=None):
         self.curdir = curdir
         self.incdirs = dirs
@@ -168,6 +170,10 @@ class IncludeDirs():
         else:
             self.extra_build_dirs = extra_build_dirs
 
+    def __repr__(self):
+        r = '<{} {}/{}>'
+        return r.format(self.__class__.__name__, self.curdir, self.incdirs)
+
     def get_curdir(self):
         return self.curdir
 
@@ -177,7 +183,7 @@ class IncludeDirs():
     def get_extra_build_dirs(self):
         return self.extra_build_dirs
 
-class ExtractedObjects():
+class ExtractedObjects:
     '''
     Holds a list of sources for which the objects must be extracted
     '''
@@ -217,7 +223,7 @@ class ExtractedObjects():
     def get_want_all_objects(self):
         return self.want_all_objects
 
-class EnvironmentVariables():
+class EnvironmentVariables:
     def __init__(self):
         self.envvars = []
 
@@ -241,7 +247,7 @@ class EnvironmentVariables():
     def prepend(self, env, name, values, kwargs):
         sep, value = self.get_value(name, values, kwargs)
         if name in env:
-            return  value + sep + env[name]
+            return value + sep + env[name]
 
         return value
 
@@ -251,11 +257,29 @@ class EnvironmentVariables():
             env[name] = method(full_env, name, values, kwargs)
         return env
 
-
-class BuildTarget():
-    def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
+class Target:
+    def __init__(self, name, subdir, build_by_default):
         self.name = name
         self.subdir = subdir
+        self.build_by_default = build_by_default
+        self.install = False
+        self.build_always = False
+
+    def get_basename(self):
+        return self.name
+
+    def get_subdir(self):
+        return self.subdir
+
+    def process_kwargs(self, kwargs):
+        if 'build_by_default' in kwargs:
+            self.build_by_default = kwargs['build_by_default']
+            if not isinstance(self.build_by_default, bool):
+                raise InvalidArguments('build_by_default must be a boolean value.')
+
+class BuildTarget(Target):
+    def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
+        super().__init__(name, subdir, True)
         self.subproject = subproject # Can not be calculated from subdir as subproject dirname can be changed per project.
         self.is_cross = is_cross
         self.is_unity = environment.coredata.get_builtin_option('unity')
@@ -288,9 +312,9 @@ class BuildTarget():
         self.process_objectlist(objects)
         self.process_kwargs(kwargs, environment)
         self.check_unknown_kwargs(kwargs)
-        if len(self.sources) == 0 and \
-            len(self.generated) == 0 and \
-            len(self.objects) == 0:
+        if len(self.sources) == 0 \
+                and len(self.generated) == 0 \
+                and len(self.objects) == 0:
             raise InvalidArguments('Build target %s has no sources.' % name)
         self.process_compilers()
         self.validate_sources()
@@ -316,7 +340,7 @@ class BuildTarget():
     def check_unknown_kwargs_int(self, kwargs, known_kwargs):
         unknowns = []
         for k in kwargs:
-            if not k in known_kwargs:
+            if k not in known_kwargs:
                 unknowns.append(k)
         if len(unknowns) > 0:
             mlog.warning('Unknown keyword argument(s) in target %s: %s.' %
@@ -330,7 +354,7 @@ class BuildTarget():
             if isinstance(s, (str, File, ExtractedObjects)):
                 self.objects.append(s)
             elif isinstance(s, (GeneratedList, CustomTarget)):
-                msg =  'Generated files are not allowed in the \'objects\' kwarg ' + \
+                msg = 'Generated files are not allowed in the \'objects\' kwarg ' + \
                     'for target {!r}.\nIt is meant only for '.format(self.name) + \
                     'pre-built object files that are shipped with the\nsource ' + \
                     'tree. Try adding it in the list of sources.'
@@ -348,7 +372,7 @@ class BuildTarget():
             if hasattr(s, 'held_object'):
                 s = s.held_object
             if isinstance(s, File):
-                if not s in added_sources:
+                if s not in added_sources:
                     self.sources.append(s)
                     added_sources[s] = True
             elif isinstance(s, (GeneratedList, CustomTarget)):
@@ -506,6 +530,7 @@ class BuildTarget():
         return self.custom_install_dir
 
     def process_kwargs(self, kwargs, environment):
+        super().process_kwargs(kwargs)
         self.copy_kwargs(kwargs)
         kwargs.get('modules', [])
         self.need_install = kwargs.get('install', self.need_install)
@@ -542,6 +567,10 @@ class BuildTarget():
         if not isinstance(valalist, list):
             valalist = [valalist]
         self.add_compiler_args('vala', valalist)
+        fortranlist = kwargs.get('fortran_args', [])
+        if not isinstance(fortranlist, list):
+            fortranlist = [fortranlist]
+        self.add_compiler_args('fortran', fortranlist)
         if not isinstance(self, Executable):
             self.vala_header = kwargs.get('vala_header', self.name + '.h')
             self.vala_vapi = kwargs.get('vala_vapi', self.name + '.vapi')
@@ -560,14 +589,16 @@ class BuildTarget():
         for i in self.link_depends:
             if not isinstance(i, str):
                 raise InvalidArguments('Link_depends arguments must be strings.')
-        inclist = kwargs.get('include_directories', [])
-        if not isinstance(inclist, list):
-            inclist = [inclist]
-        self.add_include_dirs(inclist)
         deplist = kwargs.get('dependencies', [])
         if not isinstance(deplist, list):
             deplist = [deplist]
         self.add_deps(deplist)
+        # Target-specific include dirs must be added after include dirs from
+        # internal deps (added inside self.add_deps()) to override correctly.
+        inclist = kwargs.get('include_directories', [])
+        if not isinstance(inclist, list):
+            inclist = [inclist]
+        self.add_include_dirs(inclist)
         self.custom_install_dir = kwargs.get('install_dir', None)
         if self.custom_install_dir is not None:
             if not isinstance(self.custom_install_dir, str):
@@ -638,9 +669,6 @@ class BuildTarget():
                 if not isinstance(self.pic, bool):
                     raise InvalidArguments('Argument pic to static library {!r} must be boolean'.format(self.name))
 
-    def get_subdir(self):
-        return self.subdir
-
     def get_filename(self):
         return self.filename
 
@@ -667,9 +695,6 @@ class BuildTarget():
                 transitive_deps += t.get_dependencies()
         return transitive_deps
 
-    def get_basename(self):
-        return self.name
-
     def get_source_subdir(self):
         return self.subdir
 
@@ -721,18 +746,21 @@ class BuildTarget():
             elif isinstance(dep, dependencies.Dependency):
                 self.external_deps.append(dep)
                 self.process_sourcelist(dep.get_sources())
+            elif isinstance(dep, BuildTarget):
+                raise InvalidArguments('''Tried to use a build target as a dependency.
+You probably should put it in link_with instead.''')
             else:
                 # This is a bit of a hack. We do not want Build to know anything
                 # about the interpreter so we can't import it and use isinstance.
                 # This should be reliable enough.
-                if hasattr(dep, 'subproject'):
+                if hasattr(dep, 'args_frozen'):
                     raise InvalidArguments('Tried to use subproject object as a dependency.\n'
-                            'You probably wanted to use a dependency declared in it instead.\n'
-                            'Access it by calling get_variable() on the subproject object.')
+                                           'You probably wanted to use a dependency declared in it instead.\n'
+                                           'Access it by calling get_variable() on the subproject object.')
                 raise InvalidArguments('Argument is of an unacceptable type {!r}.\nMust be '
-                        'either an external dependency (returned by find_library() or '
-                        'dependency()) or an internal dependency (returned by '
-                        'declare_dependency()).'.format(type(dep).__name__))
+                                       'either an external dependency (returned by find_library() or '
+                                       'dependency()) or an internal dependency (returned by '
+                                       'declare_dependency()).'.format(type(dep).__name__))
 
     def get_external_deps(self):
         return self.external_deps
@@ -744,7 +772,7 @@ class BuildTarget():
             if hasattr(t, 'held_object'):
                 t = t.held_object
             if not isinstance(t, (StaticLibrary, SharedLibrary)):
-                raise InvalidArguments('Link target {!r} is not library.'.format(t.name))
+                raise InvalidArguments('Link target {!r} is not library.'.format(t))
             if isinstance(self, SharedLibrary) and isinstance(t, StaticLibrary) and not t.pic:
                 msg = "Can't link non-PIC static library {!r} into shared library {!r}. ".format(t.name, self.name)
                 msg += "Use the 'pic' option to static_library to build with PIC."
@@ -794,8 +822,8 @@ class BuildTarget():
         else:
             self.extra_args[language] = args
 
-    def get_aliaslist(self):
-        return []
+    def get_aliases(self):
+        return {}
 
     def get_clike_dynamic_linker(self):
         '''
@@ -835,7 +863,7 @@ class BuildTarget():
         return False
 
 
-class Generator():
+class Generator:
     def __init__(self, args, kwargs):
         if len(args) != 1:
             raise InvalidArguments('Generator requires exactly one positional argument: the executable')
@@ -875,7 +903,7 @@ class Generator():
         for rule in outputs:
             if not isinstance(rule, str):
                 raise InvalidArguments('"output" may only contain strings.')
-            if not '@BASENAME@' in rule and not '@PLAINNAME@' in rule:
+            if '@BASENAME@' not in rule and '@PLAINNAME@' not in rule:
                 raise InvalidArguments('Every element of "output" must contain @BASENAME@ or @PLAINNAME@.')
             if '/' in rule or '\\' in rule:
                 raise InvalidArguments('"outputs" must not contain a directory separator.')
@@ -907,7 +935,18 @@ class Generator():
     def get_arglist(self):
         return self.arglist
 
-class GeneratedList():
+    def process_files(self, name, files, state, extra_args=[]):
+        output = GeneratedList(self, extra_args=extra_args)
+        for f in files:
+            if isinstance(f, str):
+                f = File.from_source_file(state.environment.source_dir, state.subdir, f)
+            elif not isinstance(f, File):
+                raise InvalidArguments('{} arguments must be strings or files not {!r}.'.format(name, f))
+            output.add_file(f)
+        return output
+
+
+class GeneratedList:
     def __init__(self, generator, extra_args=[]):
         if hasattr(generator, 'held_object'):
             generator = generator.held_object
@@ -921,7 +960,7 @@ class GeneratedList():
 
     def add_file(self, newfile):
         self.infilelist.append(newfile)
-        outfiles = self.generator.get_base_outnames(newfile)
+        outfiles = self.generator.get_base_outnames(newfile.fname)
         self.outfilelist += outfiles
         self.outmap[newfile] = outfiles
 
@@ -1025,7 +1064,7 @@ class SharedLibrary(BuildTarget):
         First we determine the filename template (self.filename_tpl), then we
         set the output filename (self.filename).
 
-        The template is needed while creating aliases (self.get_aliaslist),
+        The template is needed while creating aliases (self.get_aliases),
         which are needed while generating .so shared libraries for Linux.
 
         Besides this, there's also the import library name, which is only used
@@ -1045,7 +1084,7 @@ class SharedLibrary(BuildTarget):
         # If the user already provided the prefix and suffix to us, we don't
         # need to do any filename suffix/prefix detection.
         # NOTE: manual prefix/suffix override is currently only tested for C/C++
-        if self.prefix != None and self.suffix != None:
+        if self.prefix is not None and self.suffix is not None:
             pass
         # C# and Mono
         elif 'cs' in self.compilers:
@@ -1100,9 +1139,9 @@ class SharedLibrary(BuildTarget):
             else:
                 # No versioning, libfoo.so
                 self.filename_tpl = '{0.prefix}{0.name}.{0.suffix}'
-        if self.prefix == None:
+        if self.prefix is None:
             self.prefix = prefix
-        if self.suffix == None:
+        if self.suffix is None:
             self.suffix = suffix
         self.filename = self.filename_tpl.format(self)
 
@@ -1145,7 +1184,7 @@ class SharedLibrary(BuildTarget):
         # Visual Studio module-definitions file
         if 'vs_module_defs' in kwargs:
             path = kwargs['vs_module_defs']
-            if (os.path.isabs(path)):
+            if os.path.isabs(path):
                 self.vs_module_defs = File.from_absolute_file(path)
             else:
                 self.vs_module_defs = File.from_source_file(environment.source_dir, self.subdir, path)
@@ -1171,25 +1210,30 @@ class SharedLibrary(BuildTarget):
     def get_all_link_deps(self):
         return [self] + self.get_transitive_link_deps()
 
-    def get_aliaslist(self):
+    def get_aliases(self):
         """
         If the versioned library name is libfoo.so.0.100.0, aliases are:
-        * libfoo.so.0 (soversion)
-        * libfoo.so (unversioned; for linking)
+        * libfoo.so.0 (soversion) -> libfoo.so.0.100.0
+        * libfoo.so (unversioned; for linking) -> libfoo.so.0
         """
+        aliases = {}
         # Aliases are only useful with .so libraries. Also if the .so library
         # ends with .so (no versioning), we don't need aliases.
         if self.suffix != 'so' or self.filename.endswith('.so'):
-            return []
-        # Unversioned alias: libfoo.so
-        aliases = [self.basic_filename_tpl.format(self)]
-        # If ltversion != soversion we create an soversion alias: libfoo.so.X
+            return {}
+        # If ltversion != soversion we create an soversion alias:
+        # libfoo.so.0 -> libfoo.so.0.100.0
         if self.ltversion and self.ltversion != self.soversion:
             if not self.soversion:
                 # This is done in self.process_kwargs()
                 raise AssertionError('BUG: If library version is defined, soversion must have been defined')
             alias_tpl = self.filename_tpl.replace('ltversion', 'soversion')
-            aliases.append(alias_tpl.format(self))
+            ltversion_filename = alias_tpl.format(self)
+            aliases[ltversion_filename] = self.filename
+        else:
+            ltversion_filename = self.filename
+        # Unversioned alias: libfoo.so -> libfoo.so.0
+        aliases[self.basic_filename_tpl.format(self)] = ltversion_filename
         return aliases
 
     def type_suffix(self):
@@ -1205,22 +1249,22 @@ class SharedModule(SharedLibrary):
             raise MesonException('Shared modules must not specify the soversion kwarg.')
         super().__init__(name, subdir, subproject, is_cross, sources, objects, environment, kwargs)
 
-class CustomTarget:
-    known_kwargs = {'input' : True,
-                    'output' : True,
-                    'command' : True,
-                    'capture' : False,
-                    'install' : True,
-                    'install_dir' : True,
-                    'build_always' : True,
-                    'depends' : True,
-                    'depend_files' : True,
-                    'depfile' : True,
+class CustomTarget(Target):
+    known_kwargs = {'input': True,
+                    'output': True,
+                    'command': True,
+                    'capture': False,
+                    'install': True,
+                    'install_dir': True,
+                    'build_always': True,
+                    'depends': True,
+                    'depend_files': True,
+                    'depfile': True,
+                    'build_by_default': True,
                     }
 
     def __init__(self, name, subdir, kwargs, absolute_paths=False):
-        self.name = name
-        self.subdir = subdir
+        super().__init__(name, subdir, False)
         self.dependencies = []
         self.extra_depends = []
         self.depend_files = [] # Files that this target depends on but are not on the command line.
@@ -1255,6 +1299,7 @@ class CustomTarget:
         return deps
 
     def process_kwargs(self, kwargs):
+        super().process_kwargs(kwargs)
         self.sources = kwargs.get('input', [])
         if not isinstance(self.sources, list):
             self.sources = [self.sources]
@@ -1344,9 +1389,6 @@ class CustomTarget:
                 mlog.debug(i)
                 raise InvalidArguments('Unknown type {!r} in depend_files.'.format(type(i).__name__))
 
-    def get_basename(self):
-        return self.name
-
     def get_dependencies(self):
         return self.dependencies
 
@@ -1356,9 +1398,6 @@ class CustomTarget:
     def get_custom_install_dir(self):
         return self.install_dir
 
-    def get_subdir(self):
-        return self.subdir
-
     def get_outputs(self):
         return self.output
 
@@ -1374,13 +1413,12 @@ class CustomTarget:
     def type_suffix(self):
         return "@cus"
 
-class RunTarget:
+class RunTarget(Target):
     def __init__(self, name, command, args, dependencies, subdir):
-        self.name = name
+        super().__init__(name, subdir, False)
         self.command = command
         self.args = args
         self.dependencies = dependencies
-        self.subdir = subdir
 
     def __repr__(self):
         repr_str = "<{0} {1}: {2}>"
@@ -1389,9 +1427,6 @@ class RunTarget:
     def get_id(self):
         return self.name + self.type_suffix()
 
-    def get_basename(self):
-        return self.name
-
     def get_dependencies(self):
         return self.dependencies
 
@@ -1401,9 +1436,6 @@ class RunTarget:
     def get_sources(self):
         return []
 
-    def get_subdir(self):
-        return self.subdir
-
     def should_install(self):
         return False
 
@@ -1415,7 +1447,7 @@ class RunTarget:
 
 class Jar(BuildTarget):
     def __init__(self, name, subdir, subproject, is_cross, sources, objects, environment, kwargs):
-        super().__init__(name, subdir, subproject, is_cross, sources, objects, environment, kwargs);
+        super().__init__(name, subdir, subproject, is_cross, sources, objects, environment, kwargs)
         for s in self.sources:
             if not s.endswith('.java'):
                 raise InvalidArguments('Jar source %s is not a java file.' % s)
@@ -1431,7 +1463,7 @@ class Jar(BuildTarget):
     def get_java_args(self):
         return self.java_args
 
-class ConfigureFile():
+class ConfigureFile:
 
     def __init__(self, subdir, sourcename, targetname, configuration_data):
         self.subdir = subdir
@@ -1457,7 +1489,7 @@ class ConfigureFile():
     def get_target_name(self):
         return self.targetname
 
-class ConfigurationData():
+class ConfigurationData:
     def __init__(self):
         super().__init__()
         self.values = {}
@@ -1465,18 +1497,22 @@ class ConfigurationData():
     def __repr__(self):
         return repr(self.values)
 
+    def __contains__(self, value):
+        return value in self.values
+
     def get(self, name):
-        return self.values[name]                 # (val, desc)
+        return self.values[name] # (val, desc)
 
     def keys(self):
         return self.values.keys()
 
 # A bit poorly named, but this represents plain data files to copy
 # during install.
-class Data():
-    def __init__(self, sources, install_dir):
+class Data:
+    def __init__(self, sources, install_dir, install_mode=None):
         self.sources = sources
         self.install_dir = install_dir
+        self.install_mode = install_mode
         if not isinstance(self.sources, list):
             self.sources = [self.sources]
         for s in self.sources:
@@ -1489,3 +1525,10 @@ class RunScript(dict):
         assert(isinstance(args, list))
         self['exe'] = script
         self['args'] = args
+
+class TestSetup:
+    def __init__(self, *, exe_wrapper=None, gdb=None, timeout_multiplier=None, env=None):
+        self.exe_wrapper = exe_wrapper
+        self.gdb = gdb
+        self.timeout_multiplier = timeout_multiplier
+        self.env = env
