@@ -1,4 +1,4 @@
-# Copyright 2012-2016 The Meson development team
+# Copyright 2012-2017 The Meson development team
 
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -13,9 +13,11 @@
 # limitations under the License.
 
 import pickle, os, uuid
-from .mesonlib import MesonException, default_libdir, default_libexecdir, default_prefix
+from pathlib import PurePath
+from .mesonlib import MesonException, commonpath
+from .mesonlib import default_libdir, default_libexecdir, default_prefix
 
-version = '0.37.1'
+version = '0.38.1'
 backendlist = ['ninja', 'vs2010', 'vs2015', 'xcode']
 
 class UserOption:
@@ -43,7 +45,7 @@ class UserStringOption(UserOption):
 
 class UserBooleanOption(UserOption):
     def __init__(self, name, description, value):
-        super().__init__(name, description, [ True, False ])
+        super().__init__(name, description, [True, False])
         self.set_value(value)
 
     def tobool(self, thing):
@@ -105,7 +107,7 @@ class UserStringArrayOption(UserOption):
 # invocations of Meson. It is roughly the same thing as
 # cmakecache.
 
-class CoreData():
+class CoreData:
 
     def __init__(self, options):
         self.guid = str(uuid.uuid4()).upper()
@@ -117,7 +119,9 @@ class CoreData():
         self.user_options = {}
         self.compiler_options = {}
         self.base_options = {}
-        self.external_args = {} # These are set from "the outside" with e.g. mesonconf
+        # These two, external_*args, are set via env vars CFLAGS, LDFLAGS, etc
+        # but only when not cross-compiling.
+        self.external_args = {}
         self.external_link_args = {}
         if options.cross_file is not None:
             self.cross_file = os.path.join(os.getcwd(), options.cross_file)
@@ -131,10 +135,57 @@ class CoreData():
         # Only to print a warning if it changes between Meson invocations.
         self.pkgconf_envvar = os.environ.get('PKG_CONFIG_PATH', '')
 
+    def sanitize_prefix(self, prefix):
+        if not os.path.isabs(prefix):
+            raise MesonException('prefix value {!r} must be an absolute path'
+                                 ''.format(prefix))
+        if prefix.endswith('/') or prefix.endswith('\\'):
+            # On Windows we need to preserve the trailing slash if the
+            # string is of type 'C:\' because 'C:' is not an absolute path.
+            if len(prefix) == 3 and prefix[1] == ':':
+                pass
+            else:
+                prefix = prefix[:-1]
+        return prefix
+
+    def sanitize_dir_option_value(self, prefix, option, value):
+        '''
+        If the option is an installation directory option and the value is an
+        absolute path, check that it resides within prefix and return the value
+        as a path relative to the prefix.
+
+        This way everyone can do f.ex, get_option('libdir') and be sure to get
+        the library directory relative to prefix.
+        '''
+        if option.endswith('dir') and os.path.isabs(value) and \
+           option not in builtin_dir_noprefix_options:
+            # Value must be a subdir of the prefix
+            # commonpath will always return a path in the native format, so we
+            # must use pathlib.PurePath to do the same conversion before
+            # comparing.
+            if commonpath([value, prefix]) != str(PurePath(prefix)):
+                m = 'The value of the {!r} option is {!r} which must be a ' \
+                    'subdir of the prefix {!r}.\nNote that if you pass a ' \
+                    'relative path, it is assumed to be a subdir of prefix.'
+                raise MesonException(m.format(option, value, prefix))
+            # Convert path to be relative to prefix
+            skip = len(prefix) + 1
+            value = value[skip:]
+        return value
+
     def init_builtins(self, options):
         self.builtins = {}
+        # Sanitize prefix
+        options.prefix = self.sanitize_prefix(options.prefix)
+        # Initialize other builtin options
         for key in get_builtin_options():
-            args = [key] + builtin_options[key][1:-1] + [ getattr(options, key, get_builtin_option_default(key)) ]
+            if hasattr(options, key):
+                value = getattr(options, key)
+                value = self.sanitize_dir_option_value(options.prefix, key, value)
+                setattr(options, key, value)
+            else:
+                value = get_builtin_option_default(key)
+            args = [key] + builtin_options[key][1:-1] + [value]
             self.builtins[key] = builtin_options[key][0](*args)
 
     def get_builtin_option(self, optname):
@@ -143,10 +194,14 @@ class CoreData():
         raise RuntimeError('Tried to get unknown builtin option %s.' % optname)
 
     def set_builtin_option(self, optname, value):
-        if optname in self.builtins:
-            self.builtins[optname].set_value(value)
+        if optname == 'prefix':
+            value = self.sanitize_prefix(value)
+        elif optname in self.builtins:
+            prefix = self.builtins['prefix'].value
+            value = self.sanitize_dir_option_value(prefix, optname, value)
         else:
             raise RuntimeError('Tried to set unknown builtin option %s.' % optname)
+        self.builtins[optname].set_value(value)
 
 def load(filename):
     load_fail_msg = 'Coredata file {!r} is corrupted. Try with a fresh build tree.'.format(filename)
@@ -158,8 +213,8 @@ def load(filename):
     if not isinstance(obj, CoreData):
         raise MesonException(load_fail_msg)
     if obj.version != version:
-        raise MesonException('Build directory has been generated with Meson version %s, which is incompatible with current version %s.\nPlease delete this build directory AND create a new one.'%
-                           (obj.version, version))
+        raise MesonException('Build directory has been generated with Meson version %s, which is incompatible with current version %s.\nPlease delete this build directory AND create a new one.' %
+                             (obj.version, version))
     return obj
 
 def save(obj, filename):
@@ -179,7 +234,7 @@ def get_builtin_option_choices(optname):
         if builtin_options[optname][0] == UserStringOption:
             return None
         elif builtin_options[optname][0] == UserBooleanOption:
-            return [ True, False ]
+            return [True, False]
         else:
             return builtin_options[optname][2]
     else:
@@ -201,19 +256,19 @@ def get_builtin_option_default(optname):
         raise RuntimeError('Tried to get the default value for an unknown builtin option \'%s\'.' % optname)
 
 builtin_options = {
-        'buildtype'         : [ UserComboOption, 'Build type to use.', [ 'plain', 'debug', 'debugoptimized', 'release', 'minsize' ], 'debug' ],
-        'strip'             : [ UserBooleanOption, 'Strip targets on install.', False ],
-        'unity'             : [ UserBooleanOption, 'Unity build.', False ],
-        'prefix'            : [ UserStringOption, 'Installation prefix.', default_prefix() ],
-        'libdir'            : [ UserStringOption, 'Library directory.', default_libdir() ],
-        'libexecdir'        : [ UserStringOption, 'Library executable directory.', default_libexecdir() ],
-        'bindir'            : [ UserStringOption, 'Executable directory.', 'bin' ],
-        'sbindir'           : [ UserStringOption, 'System executable directory.', 'sbin' ],
-        'includedir'        : [ UserStringOption, 'Header file directory.', 'include' ],
-        'datadir'           : [ UserStringOption, 'Data file directory.', 'share' ],
-        'mandir'            : [ UserStringOption, 'Manual page directory.', 'share/man' ],
-        'infodir'           : [ UserStringOption, 'Info page directory.', 'share/info' ],
-        'localedir'         : [ UserStringOption, 'Locale data directory.', 'share/locale' ],
+    'buildtype':  [UserComboOption, 'Build type to use.', ['plain', 'debug', 'debugoptimized', 'release', 'minsize'], 'debug'],
+    'strip':      [UserBooleanOption, 'Strip targets on install.', False],
+    'unity':      [UserBooleanOption, 'Unity build.', False],
+    'prefix':     [UserStringOption, 'Installation prefix.', default_prefix()],
+    'libdir':     [UserStringOption, 'Library directory.', default_libdir()],
+    'libexecdir': [UserStringOption, 'Library executable directory.', default_libexecdir()],
+    'bindir':     [UserStringOption, 'Executable directory.', 'bin'],
+    'sbindir':    [UserStringOption, 'System executable directory.', 'sbin'],
+    'includedir': [UserStringOption, 'Header file directory.', 'include'],
+    'datadir':    [UserStringOption, 'Data file directory.', 'share'],
+    'mandir':     [UserStringOption, 'Manual page directory.', 'share/man'],
+    'infodir':    [UserStringOption, 'Info page directory.', 'share/info'],
+    'localedir':  [UserStringOption, 'Locale data directory.', 'share/locale'],
     # sysconfdir, localstatedir and sharedstatedir are a bit special. These defaults to ${prefix}/etc,
     # ${prefix}/var and ${prefix}/com but nobody uses that. Instead they always set it
     # manually to /etc, /var and /var/lib. This default values is thus pointless and not really used
@@ -223,17 +278,20 @@ builtin_options = {
     # to set the following in project():
     #
     # default_options : ['sysconfdir=/etc', 'localstatedir=/var', 'sharedstatedir=/var/lib']
-        'sysconfdir'        : [ UserStringOption, 'Sysconf data directory.', 'etc' ],
-        'localstatedir'     : [ UserStringOption, 'Localstate data directory.', 'var' ],
-        'sharedstatedir'    : [ UserStringOption, 'Architecture-independent data directory.', 'com' ],
-        'werror'            : [ UserBooleanOption, 'Treat warnings as errors.', False ],
-        'warning_level'     : [ UserComboOption, 'Compiler warning level to use.', [ '1', '2', '3' ], '1'],
-        'layout'            : [ UserComboOption, 'Build directory layout.', ['mirror', 'flat' ], 'mirror' ],
-        'default_library'   : [ UserComboOption, 'Default library type.', [ 'shared', 'static' ], 'shared' ],
-        'backend'           : [ UserComboOption, 'Backend to use.', backendlist, 'ninja' ],
-        'stdsplit'          : [ UserBooleanOption, 'Split stdout and stderr in test logs.', True ],
-        'errorlogs'         : [ UserBooleanOption, "Whether to print the logs from failing tests.", True ],
-        }
+    'sysconfdir':      [UserStringOption, 'Sysconf data directory.', 'etc'],
+    'localstatedir':   [UserStringOption, 'Localstate data directory.', 'var'],
+    'sharedstatedir':  [UserStringOption, 'Architecture-independent data directory.', 'com'],
+    'werror':          [UserBooleanOption, 'Treat warnings as errors.', False],
+    'warning_level':   [UserComboOption, 'Compiler warning level to use.', ['1', '2', '3'], '1'],
+    'layout':          [UserComboOption, 'Build directory layout.', ['mirror', 'flat'], 'mirror'],
+    'default_library': [UserComboOption, 'Default library type.', ['shared', 'static'], 'shared'],
+    'backend':         [UserComboOption, 'Backend to use.', backendlist, 'ninja'],
+    'stdsplit':        [UserBooleanOption, 'Split stdout and stderr in test logs.', True],
+    'errorlogs':       [UserBooleanOption, "Whether to print the logs from failing tests.", True],
+}
+
+# Installation directories that can reside in a path outside of the prefix
+builtin_dir_noprefix_options = {'sysconfdir', 'localstatedir', 'sharedstatedir'}
 
 forbidden_target_names = {'clean': None,
                           'clean-ctlist': None,
@@ -249,7 +307,8 @@ forbidden_target_names = {'clean': None,
                           'test:': None,
                           'benchmark': None,
                           'install': None,
+                          'uninstall': None,
                           'build.ninja': None,
                           'scan-build': None,
                           'reconfigure': None,
-                         }
+                          }
