@@ -20,7 +20,9 @@
 # package before this gets too big.
 
 import re
-import os, stat, glob, subprocess, shutil
+import sys
+import os, stat, glob, shutil
+import subprocess
 import sysconfig
 from collections import OrderedDict
 from . mesonlib import MesonException, version_compare, version_compare_many, Popen_safe
@@ -29,10 +31,9 @@ from . import mesonlib
 from .environment import detect_cpu_family, for_windows
 
 class DependencyException(MesonException):
-    def __init__(self, *args, **kwargs):
-        MesonException.__init__(self, *args, **kwargs)
+    '''Exceptions raised while trying to find dependencies'''
 
-class Dependency():
+class Dependency:
     def __init__(self, type_name='unknown'):
         self.name = "null"
         self.is_found = False
@@ -59,7 +60,7 @@ class Dependency():
     def get_name(self):
         return self.name
 
-    def get_exe_args(self):
+    def get_exe_args(self, compiler):
         return []
 
     def need_threads(self):
@@ -89,7 +90,9 @@ class InternalDependency(Dependency):
         return self.version
 
 class PkgConfigDependency(Dependency):
-    pkgbin = None
+    # The class's copy of the pkg-config path. Avoids having to search for it
+    # multiple times in the same Meson invocation.
+    class_pkgbin = None
 
     def __init__(self, name, environment, kwargs):
         Dependency.__init__(self, 'pkgconfig')
@@ -99,33 +102,47 @@ class PkgConfigDependency(Dependency):
         self.silent = kwargs.get('silent', False)
         if not isinstance(self.static, bool):
             raise DependencyException('Static keyword must be boolean')
+        # Store a copy of the pkg-config path on the object itself so it is
+        # stored in the pickled coredata and recovered.
+        self.pkgbin = None
         self.cargs = []
         self.libs = []
         if 'native' in kwargs and environment.is_cross_build():
-            want_cross = not kwargs['native']
+            self.want_cross = not kwargs['native']
         else:
-            want_cross = environment.is_cross_build()
+            self.want_cross = environment.is_cross_build()
         self.name = name
 
         # When finding dependencies for cross-compiling, we don't care about
         # the 'native' pkg-config
-        if want_cross:
+        if self.want_cross:
             if 'pkgconfig' not in environment.cross_info.config['binaries']:
                 if self.required:
                     raise DependencyException('Pkg-config binary missing from cross file')
             else:
-                PkgConfigDependency.pkgbin = environment.cross_info.config['binaries']['pkgconfig']
+                potential_pkgbin = ExternalProgram(environment.cross_info.config['binaries'].get('pkgconfig', 'non_existing_binary'),
+                                                   silent=True)
+                if potential_pkgbin.found():
+                    # FIXME, we should store all pkg-configs in ExternalPrograms.
+                    # However that is too destabilizing a change to do just before release.
+                    self.pkgbin = potential_pkgbin.get_command()[0]
+                    PkgConfigDependency.class_pkgbin = self.pkgbin
+                else:
+                    mlog.debug('Cross pkg-config %s not found.' % potential_pkgbin.name)
         # Only search for the native pkg-config the first time and
         # store the result in the class definition
-        elif PkgConfigDependency.pkgbin is None:
-            PkgConfigDependency.pkgbin = self.check_pkgconfig()
+        elif PkgConfigDependency.class_pkgbin is None:
+            self.pkgbin = self.check_pkgconfig()
+            PkgConfigDependency.class_pkgbin = self.pkgbin
+        else:
+            self.pkgbin = PkgConfigDependency.class_pkgbin
 
         self.is_found = False
         if not self.pkgbin:
             if self.required:
                 raise DependencyException('Pkg-config not found.')
             return
-        if want_cross:
+        if self.want_cross:
             self.type_string = 'Cross'
         else:
             self.type_string = 'Native'
@@ -160,17 +177,18 @@ class PkgConfigDependency(Dependency):
                 if not self.silent:
                     mlog.log(*found_msg)
                 if self.required:
-                    raise DependencyException(
-                        'Invalid version of a dependency, needed %s %s found %s.' %
-                        (name, not_found, self.modversion))
+                    m = 'Invalid version of dependency, need {!r} {!r} found {!r}.'
+                    raise DependencyException(m.format(name, not_found, self.modversion))
                 return
         found_msg += [mlog.green('YES'), self.modversion]
-        if not self.silent:
-            mlog.log(*found_msg)
         # Fetch cargs to be used while using this dependency
         self._set_cargs()
         # Fetch the libraries and library paths needed for using this
         self._set_libs()
+        # Print the found message only at the very end because fetching cflags
+        # and libs can also fail if other needed pkg-config files aren't found.
+        if not self.silent:
+            mlog.log(*found_msg)
 
     def __repr__(self):
         s = '<{0} {1}: {2} {3}>'
@@ -179,12 +197,12 @@ class PkgConfigDependency(Dependency):
 
     def _call_pkgbin(self, args):
         p, out = Popen_safe([self.pkgbin] + args, env=os.environ)[0:2]
-        return (p.returncode, out.strip())
+        return p.returncode, out.strip()
 
     def _set_cargs(self):
         ret, out = self._call_pkgbin(['--cflags', self.name])
         if ret != 0:
-            raise DependencyException('Could not generate cargs for %s:\n\n%s' % \
+            raise DependencyException('Could not generate cargs for %s:\n\n%s' %
                                       (self.name, out))
         self.cargs = out.split()
 
@@ -194,7 +212,7 @@ class PkgConfigDependency(Dependency):
             libcmd.append('--static')
         ret, out = self._call_pkgbin(libcmd)
         if ret != 0:
-            raise DependencyException('Could not generate libs for %s:\n\n%s' % \
+            raise DependencyException('Could not generate libs for %s:\n\n%s' %
                                       (self.name, out))
         self.libs = []
         for lib in out.split():
@@ -228,7 +246,7 @@ class PkgConfigDependency(Dependency):
         return self.modversion
 
     def get_version(self):
-        return self.get_modversion()
+        return self.modversion
 
     def get_compile_args(self):
         return self.cargs
@@ -306,6 +324,7 @@ class WxDependency(Dependency):
     def __init__(self, environment, kwargs):
         Dependency.__init__(self, 'wx')
         self.is_found = False
+        self.modversion = 'none'
         if WxDependency.wx_found is None:
             self.check_wxconfig()
         if not WxDependency.wx_found:
@@ -322,7 +341,7 @@ class WxDependency(Dependency):
             version_req = kwargs.get('version', None)
             if version_req is not None:
                 if not version_compare(self.modversion, version_req, strict=True):
-                    mlog.log('Wxwidgets version %s does not fullfill requirement %s' %\
+                    mlog.log('Wxwidgets version %s does not fullfill requirement %s' %
                              (self.modversion, version_req))
                     return
             mlog.log('Dependency wxwidgets found:', mlog.green('YES'))
@@ -342,7 +361,7 @@ class WxDependency(Dependency):
 
     def get_requested(self, kwargs):
         modules = 'modules'
-        if not modules in kwargs:
+        if modules not in kwargs:
             return []
         candidates = kwargs[modules]
         if isinstance(candidates, str):
@@ -355,6 +374,9 @@ class WxDependency(Dependency):
     def get_modversion(self):
         return self.modversion
 
+    def get_version(self):
+        return self.modversion
+
     def get_compile_args(self):
         return self.cargs
 
@@ -379,7 +401,7 @@ class WxDependency(Dependency):
     def found(self):
         return self.is_found
 
-class ExternalProgram():
+class ExternalProgram:
     windows_exts = ('exe', 'com', 'bat')
 
     def __init__(self, name, fullpath=None, silent=False, search_dir=None):
@@ -410,10 +432,15 @@ class ExternalProgram():
             if first_line.startswith('#!'):
                 commands = first_line[2:].split('#')[0].strip().split()
                 if mesonlib.is_windows():
-                    # Windows does not have /usr/bin.
-                    commands[0] = commands[0].split('/')[-1]
-                    if commands[0] == 'env':
+                    # Windows does not have UNIX paths so remove them,
+                    # but don't remove Windows paths
+                    if commands[0].startswith('/'):
+                        commands[0] = commands[0].split('/')[-1]
+                    if len(commands) > 0 and commands[0] == 'env':
                         commands = commands[1:]
+                    # Windows does not ship python3.exe, but we know the path to it
+                    if len(commands) > 0 and commands[0] == 'python3':
+                        commands[0] = sys.executable
                 return commands + [script]
         except Exception:
             pass
@@ -516,7 +543,7 @@ class BoostDependency(Dependency):
     # Some boost libraries have different names for
     # their sources and libraries. This dict maps
     # between the two.
-    name2lib = {'test' : 'unit_test_framework'}
+    name2lib = {'test': 'unit_test_framework'}
 
     def __init__(self, environment, kwargs):
         Dependency.__init__(self, 'boost')
@@ -524,9 +551,9 @@ class BoostDependency(Dependency):
         self.environment = environment
         self.libdir = ''
         if 'native' in kwargs and environment.is_cross_build():
-            want_cross = not kwargs['native']
+            self.want_cross = not kwargs['native']
         else:
-            want_cross = environment.is_cross_build()
+            self.want_cross = environment.is_cross_build()
         try:
             self.boost_root = os.environ['BOOST_ROOT']
             if not os.path.isabs(self.boost_root):
@@ -534,7 +561,7 @@ class BoostDependency(Dependency):
         except KeyError:
             self.boost_root = None
         if self.boost_root is None:
-            if want_cross:
+            if self.want_cross:
                 raise DependencyException('BOOST_ROOT is needed while cross-compiling')
             if mesonlib.is_windows():
                 self.boost_root = self.detect_win_root()
@@ -559,8 +586,7 @@ class BoostDependency(Dependency):
                 info = self.version + ', ' + self.boost_root
             else:
                 info = self.version
-            mlog.log('Dependency Boost (%s) found:' % module_str, mlog.green('YES'),
-                     '(' + info + ')')
+            mlog.log('Dependency Boost (%s) found:' % module_str, mlog.green('YES'), info)
         else:
             mlog.log("Dependency Boost (%s) found:" % module_str, mlog.red('NO'))
 
@@ -656,7 +682,6 @@ class BoostDependency(Dependency):
             self.lib_modules_mt[modname] = fname
 
     def detect_lib_modules_nix(self):
-        libsuffix = None
         if mesonlib.is_osx():
             libsuffix = 'dylib'
         else:
@@ -785,8 +810,10 @@ class GTestDependency(Dependency):
 
     def get_link_args(self):
         return self.link_args
+
     def get_version(self):
         return '1.something_maybe'
+
     def get_sources(self):
         return self.sources
 
@@ -953,7 +980,7 @@ class QtBaseDependency(Dependency):
             if not self.qmake.found():
                 continue
             # Check that the qmake is for qt5
-            pc, stdo = Popen_safe(self.qmake.fullpath +  ['-v'])[0:2]
+            pc, stdo = Popen_safe(self.qmake.fullpath + ['-v'])[0:2]
             if pc.returncode != 0:
                 continue
             if not 'Qt version ' + self.qtver in stdo:
@@ -966,7 +993,7 @@ class QtBaseDependency(Dependency):
             return
         self.version = re.search(self.qtver + '(\.\d+)+', stdo).group(0)
         # Query library path, header path, and binary path
-        stdo = Popen_safe(self.qmake.fullpath +  ['-query'])[1]
+        stdo = Popen_safe(self.qmake.fullpath + ['-query'])[1]
         qvars = {}
         for line in stdo.split('\n'):
             line = line.strip()
@@ -1029,16 +1056,14 @@ class QtBaseDependency(Dependency):
     def found(self):
         return self.is_found
 
-    def get_exe_args(self):
+    def get_exe_args(self, compiler):
         # Originally this was -fPIE but nowadays the default
         # for upstream and distros seems to be -reduce-relocations
         # which requires -fPIC. This may cause a performance
         # penalty when using self-built Qt or on platforms
         # where -fPIC is not required. If this is an issue
         # for you, patches are welcome.
-        if mesonlib.is_linux():
-            return ['-fPIC']
-        return []
+        return compiler.get_pic_args()
 
 class Qt5Dependency(QtBaseDependency):
     def __init__(self, env, kwargs):
@@ -1051,13 +1076,14 @@ class Qt4Dependency(QtBaseDependency):
 class GnuStepDependency(Dependency):
     def __init__(self, environment, kwargs):
         Dependency.__init__(self, 'gnustep')
+        self.required = kwargs.get('required', True)
         self.modules = kwargs.get('modules', [])
         self.detect()
 
     def detect(self):
-        confprog = 'gnustep-config'
+        self.confprog = 'gnustep-config'
         try:
-            gp = Popen_safe([confprog, '--help'])[0]
+            gp = Popen_safe([self.confprog, '--help'])[0]
         except (FileNotFoundError, PermissionError):
             self.args = None
             mlog.log('Dependency GnuStep found:', mlog.red('NO'), '(no gnustep-config)')
@@ -1070,16 +1096,18 @@ class GnuStepDependency(Dependency):
             arg = '--gui-libs'
         else:
             arg = '--base-libs'
-        fp, flagtxt, flagerr = Popen_safe([confprog, '--objc-flags'])
+        fp, flagtxt, flagerr = Popen_safe([self.confprog, '--objc-flags'])
         if fp.returncode != 0:
             raise DependencyException('Error getting objc-args: %s %s' % (flagtxt, flagerr))
         args = flagtxt.split()
         self.args = self.filter_arsg(args)
-        fp, libtxt, liberr = Popen_safe([confprog, arg])
+        fp, libtxt, liberr = Popen_safe([self.confprog, arg])
         if fp.returncode != 0:
             raise DependencyException('Error getting objc-lib args: %s %s' % (libtxt, liberr))
         self.libs = self.weird_filter(libtxt.split())
-        mlog.log('Dependency GnuStep found:', mlog.green('YES'))
+        self.version = self.detect_version()
+        mlog.log('Dependency', mlog.bold('GnuStep'), 'found:',
+                 mlog.green('YES'), self.version)
 
     def weird_filter(self, elems):
         """When building packages, the output of the enclosing Make
@@ -1087,21 +1115,54 @@ is sometimes mixed among the subprocess output. I have no idea
 why. As a hack filter out everything that is not a flag."""
         return [e for e in elems if e.startswith('-')]
 
-
     def filter_arsg(self, args):
         """gnustep-config returns a bunch of garbage args such
         as -O2 and so on. Drop everything that is not needed."""
         result = []
         for f in args:
-            if f.startswith('-D') or f.startswith('-f') or \
-            f.startswith('-I') or f == '-pthread' or\
-            (f.startswith('-W') and not f == '-Wall'):
+            if f.startswith('-D') \
+                    or f.startswith('-f') \
+                    or f.startswith('-I') \
+                    or f == '-pthread' \
+                    or (f.startswith('-W') and not f == '-Wall'):
                 result.append(f)
         return result
 
+    def detect_version(self):
+        gmake = self.get_variable('GNUMAKE')
+        makefile_dir = self.get_variable('GNUSTEP_MAKEFILES')
+        # This Makefile has the GNUStep version set
+        base_make = os.path.join(makefile_dir, 'Additional', 'base.make')
+        # Print the Makefile variable passed as the argument. For instance, if
+        # you run the make target `print-SOME_VARIABLE`, this will print the
+        # value of the variable `SOME_VARIABLE`.
+        printver = "print-%:\n\t@echo '$($*)'"
+        env = os.environ.copy()
+        # See base.make to understand why this is set
+        env['FOUNDATION_LIB'] = 'gnu'
+        p, o, e = Popen_safe([gmake, '-f', '-', '-f', base_make,
+                              'print-GNUSTEP_BASE_VERSION'],
+                             env=env, write=printver, stdin=subprocess.PIPE)
+        version = o.strip()
+        if not version:
+            mlog.debug("Couldn't detect GNUStep version, falling back to '1'")
+            # Fallback to setting some 1.x version
+            version = '1'
+        return version
+
+    def get_variable(self, var):
+        p, o, e = Popen_safe([self.confprog, '--variable=' + var])
+        if p.returncode != 0 and self.required:
+            raise DependencyException('{!r} for variable {!r} failed to run'
+                                      ''.format(self.confprog, var))
+        return o.strip()
+
     def found(self):
         return self.args is not None
 
+    def get_version(self):
+        return self.version
+
     def get_compile_args(self):
         if self.args is None:
             return []
@@ -1130,6 +1191,9 @@ class AppleFrameworks(Dependency):
     def found(self):
         return mesonlib.is_osx()
 
+    def get_version(self):
+        return 'unknown'
+
 class GLDependency(Dependency):
     def __init__(self, environment, kwargs):
         Dependency.__init__(self, 'gl')
@@ -1143,21 +1207,27 @@ class GLDependency(Dependency):
                 self.is_found = True
                 self.cargs = pcdep.get_compile_args()
                 self.linkargs = pcdep.get_link_args()
+                self.version = pcdep.get_version()
                 return
         except Exception:
             pass
         if mesonlib.is_osx():
             self.is_found = True
             self.linkargs = ['-framework', 'OpenGL']
+            self.version = '1' # FIXME
             return
         if mesonlib.is_windows():
             self.is_found = True
             self.linkargs = ['-lopengl32']
+            self.version = '1' # FIXME: unfixable?
             return
 
     def get_link_args(self):
         return self.linkargs
 
+    def get_version(self):
+        return self.version
+
 # There are three different ways of depending on SDL2:
 # sdl2-config, pkg-config and OSX framework
 class SDL2Dependency(Dependency):
@@ -1180,13 +1250,15 @@ class SDL2Dependency(Dependency):
             pass
         sdlconf = shutil.which('sdl2-config')
         if sdlconf:
-            pc, stdo = Popen_safe(['sdl2-config', '--cflags'])[0:2]
+            stdo = Popen_safe(['sdl2-config', '--cflags'])[1]
             self.cargs = stdo.strip().split()
-            pc, stdo = Popen_safe(['sdl2-config', '--libs'])[0:2]
+            stdo = Popen_safe(['sdl2-config', '--libs'])[1]
             self.linkargs = stdo.strip().split()
+            stdo = Popen_safe(['sdl2-config', '--version'])[1]
+            self.version = stdo.strip()
             self.is_found = True
-            mlog.log('Dependency', mlog.bold('sdl2'), 'found:', mlog.green('YES'), '(%s)' % sdlconf)
-            self.version = '2' # FIXME
+            mlog.log('Dependency', mlog.bold('sdl2'), 'found:', mlog.green('YES'),
+                     self.version, '(%s)' % sdlconf)
             return
         mlog.debug('Could not find sdl2-config binary, trying next.')
         if mesonlib.is_osx():
@@ -1252,6 +1324,9 @@ class ExtraFrameworkDependency(Dependency):
     def found(self):
         return self.name is not None
 
+    def get_version(self):
+        return 'unknown'
+
 class ThreadDependency(Dependency):
     def __init__(self, environment, kwargs):
         super().__init__('threads')
@@ -1262,12 +1337,16 @@ class ThreadDependency(Dependency):
     def need_threads(self):
         return True
 
+    def get_version(self):
+        return 'unknown'
+
 class Python3Dependency(Dependency):
     def __init__(self, environment, kwargs):
         super().__init__('python3')
         self.name = 'python3'
         self.is_found = False
-        self.version = "3.something_maybe"
+        # We can only be sure that it is Python 3 at this point
+        self.version = '3'
         try:
             pkgdep = PkgConfigDependency('python3', environment, kwargs)
             if pkgdep.found():
@@ -1377,7 +1456,9 @@ def find_external_dependency(name, environment, kwargs):
     if mesonlib.is_osx():
         fwdep = ExtraFrameworkDependency(name, required)
         if required and not fwdep.found():
-            raise DependencyException('Dependency "%s" not found' % name)
+            m = 'Dependency {!r} not found, tried Extra Frameworks ' \
+                'and Pkg-Config:\n\n' + str(pkg_exc)
+            raise DependencyException(m.format(name))
         return fwdep
     if pkg_exc is not None:
         raise pkg_exc
@@ -1393,9 +1474,9 @@ packages = {'boost': BoostDependency,
             'qt4': Qt4Dependency,
             'gnustep': GnuStepDependency,
             'appleframeworks': AppleFrameworks,
-            'wxwidgets' : WxDependency,
-            'sdl2' : SDL2Dependency,
-            'gl' : GLDependency,
-            'threads' : ThreadDependency,
-            'python3' : Python3Dependency,
-           }
+            'wxwidgets': WxDependency,
+            'sdl2': SDL2Dependency,
+            'gl': GLDependency,
+            'threads': ThreadDependency,
+            'python3': Python3Dependency,
+            }
