@@ -34,18 +34,18 @@ def find_coverage_tools():
         lcov_exe = None
     if not mesonlib.exe_exists([genhtml_exe, '--version']):
         genhtml_exe = None
-    return (gcovr_exe, lcov_exe, genhtml_exe)
+    return gcovr_exe, lcov_exe, genhtml_exe
 
-def detect_ninja():
+def detect_ninja(version='1.5'):
     for n in ['ninja', 'ninja-build']:
         try:
-            p, version = Popen_safe([n, '--version'])[0:2]
+            p, found = Popen_safe([n, '--version'])[0:2]
         except (FileNotFoundError, PermissionError):
             # Doesn't exist in PATH or isn't executable
             continue
         # Perhaps we should add a way for the caller to know the failure mode
         # (not found or too old)
-        if p.returncode == 0 and mesonlib.version_compare(version, ">=1.6"):
+        if p.returncode == 0 and mesonlib.version_compare(found, '>=' + version):
             return n
 
 def detect_native_windows_arch():
@@ -182,7 +182,7 @@ def search_version(text):
         return match.group(0)
     return 'unknown version'
 
-class Environment():
+class Environment:
     private_dir = 'meson-private'
     log_dir = 'meson-logs'
     coredata_file = os.path.join(private_dir, 'coredata.dat')
@@ -231,7 +231,7 @@ class Environment():
         # Versioning is added to these names in the backends as-needed.
         cross = self.is_cross_build()
         if (not cross and mesonlib.is_windows()) \
-        or (cross and self.cross_info.has_host() and self.cross_info.config['host_machine']['system'] == 'windows'):
+                or (cross and self.cross_info.has_host() and self.cross_info.config['host_machine']['system'] == 'windows'):
             self.exe_suffix = 'exe'
             self.object_suffix = 'obj'
             self.win_libdir_layout = True
@@ -326,6 +326,7 @@ class Environment():
             if len(rest) == 2:
                 defines[rest[0]] = rest[1]
         return defines
+
     @staticmethod
     def get_gnu_version_from_defines(defines):
         dot = '.'
@@ -386,7 +387,7 @@ class Environment():
                 version = self.get_gnu_version_from_defines(defines)
                 return GnuCCompiler(ccache + [compiler], version, gtype, is_cross, exe_wrap, defines)
             if 'clang' in out:
-                if 'Apple' in out:
+                if 'Apple' in out or for_darwin(want_cross, self):
                     cltype = CLANG_OSX
                 else:
                     cltype = CLANG_STANDARD
@@ -396,6 +397,10 @@ class Environment():
                 # everything else to stdout. Why? Lord only knows.
                 version = search_version(err)
                 return VisualStudioCCompiler([compiler], version, is_cross, exe_wrap)
+            if '(ICC)' in out:
+                # TODO: add microsoft add check OSX
+                inteltype = ICC_STANDARD
+                return IntelCCompiler(ccache + [compiler], version, inteltype, is_cross, exe_wrap)
         errmsg = 'Unknown compiler(s): "' + ', '.join(compilers) + '"'
         if popen_exceptions:
             errmsg += '\nThe follow exceptions were encountered:'
@@ -525,6 +530,10 @@ class Environment():
             if 'Microsoft' in out or 'Microsoft' in err:
                 version = search_version(err)
                 return VisualStudioCPPCompiler([compiler], version, is_cross, exe_wrap)
+            if '(ICC)' in out:
+                # TODO: add microsoft add check OSX
+                inteltype = ICC_STANDARD
+                return IntelCPPCompiler(ccache + [compiler], version, inteltype, is_cross, exe_wrap)
         errmsg = 'Unknown compiler(s): "' + ', '.join(compilers) + '"'
         if popen_exceptions:
             errmsg += '\nThe follow exceptions were encountered:'
@@ -626,7 +635,6 @@ class Environment():
         raise EnvironmentException('Unknown compiler "' + ' '.join(exelist) + '"')
 
     def detect_d_compiler(self, want_cross):
-        exelist = None
         is_cross = False
         # Search for a D compiler.
         # We prefer LDC over GDC unless overridden with the DC
@@ -680,7 +688,7 @@ class Environment():
             if evar in os.environ:
                 linker = os.environ[evar].strip()
             elif isinstance(compiler, VisualStudioCCompiler):
-                linker= self.vs_static_linker
+                linker = self.vs_static_linker
             else:
                 linker = self.default_static_linker
         basename = os.path.basename(linker).lower()
@@ -790,12 +798,15 @@ def get_args_from_envvars(compiler):
         compiler_is_linker = (compiler.get_exelist() == compiler.get_linker_exelist())
 
     if lang not in ('c', 'cpp', 'objc', 'objcpp', 'fortran', 'd'):
-        return ([], [])
+        return [], []
 
     # Compile flags
-    cflags_mapping = {'c': 'CFLAGS', 'cpp': 'CXXFLAGS',
-        'objc': 'OBJCFLAGS', 'objcpp': 'OBJCXXFLAGS',
-        'fortran': 'FFLAGS', 'd': 'DFLAGS'}
+    cflags_mapping = {'c': 'CFLAGS',
+                      'cpp': 'CXXFLAGS',
+                      'objc': 'OBJCFLAGS',
+                      'objcpp': 'OBJCXXFLAGS',
+                      'fortran': 'FFLAGS',
+                      'd': 'DFLAGS'}
     compile_flags = os.environ.get(cflags_mapping[lang], '')
     log_var(cflags_mapping[lang], compile_flags)
     compile_flags = compile_flags.split()
@@ -818,17 +829,17 @@ def get_args_from_envvars(compiler):
     log_var('CPPFLAGS', preproc_flags)
     compile_flags += preproc_flags.split()
 
-    return (compile_flags, link_flags)
+    return compile_flags, link_flags
 
-class CrossBuildInfo():
+class CrossBuildInfo:
     def __init__(self, filename):
         self.config = {'properties': {}}
         self.parse_datafile(filename)
         if 'target_machine' in self.config:
             return
-        if not 'host_machine' in self.config:
+        if 'host_machine' not in self.config:
             raise mesonlib.MesonException('Cross info file must have either host or a target machine.')
-        if not 'binaries' in self.config:
+        if 'binaries' not in self.config:
             raise mesonlib.MesonException('Cross file is missing "binaries".')
 
     def ok_type(self, i):
@@ -845,7 +856,7 @@ class CrossBuildInfo():
                 if ' ' in entry or '\t' in entry or "'" in entry or '"' in entry:
                     raise EnvironmentException('Malformed variable name %s in cross file..' % entry)
                 try:
-                    res = eval(value, {'true' : True, 'false' : False})
+                    res = eval(value, {'true': True, 'false': False})
                 except Exception:
                     raise EnvironmentException('Malformed value in cross file variable %s.' % entry)
                 if self.ok_type(res):
@@ -888,3 +899,11 @@ class CrossBuildInfo():
            self.config['host_machine']['system'] == detect_system():
             return False
         return True
+
+
+class MachineInfo:
+    def __init__(self, system, cpu_family, cpu, endian):
+        self.system = system
+        self.cpu_family = cpu_family
+        self.cpu = cpu
+        self.endian = endian
