@@ -20,13 +20,17 @@ import os
 import sys
 import copy
 import subprocess
+from . import ModuleReturnValue
 from ..mesonlib import MesonException, Popen_safe
 from ..dependencies import Dependency, PkgConfigDependency, InternalDependency
 from .. import mlog
 from .. import mesonlib
 from .. import compilers
 from .. import interpreter
-from . import find_program, GResourceTarget, GResourceHeaderTarget, GirTarget, TypelibTarget, VapiTarget
+from . import GResourceTarget, GResourceHeaderTarget, GirTarget, TypelibTarget, VapiTarget
+from . import find_program, get_include_args
+from . import ExtensionModule
+
 
 # gresource compilation is broken due to the way
 # the resource compiler and Ninja clash about it
@@ -54,7 +58,7 @@ def gir_has_extra_lib_arg():
         pass
     return _gir_has_extra_lib_arg
 
-class GnomeModule:
+class GnomeModule(ExtensionModule):
     gir_dep = None
 
     @staticmethod
@@ -161,7 +165,7 @@ can not be used with the current version of glib-compiled-resources, due to
         target_c = GResourceTarget(name, state.subdir, kwargs)
 
         if gresource: # Only one target for .gresource files
-            return [target_c]
+            return ModuleReturnValue(target_c, [target_c])
 
         h_kwargs = {
             'command': cmd,
@@ -175,7 +179,8 @@ can not be used with the current version of glib-compiled-resources, due to
             h_kwargs['install_dir'] = kwargs.get('install_dir',
                                                  state.environment.coredata.get_builtin_option('includedir'))
         target_h = GResourceHeaderTarget(args[0] + '_h', state.subdir, h_kwargs)
-        return [target_c, target_h]
+        rv = [target_c, target_h]
+        return ModuleReturnValue(rv, rv)
 
     def _get_gresource_dependencies(self, state, input_file, source_dirs, dependencies):
         for dep in dependencies:
@@ -250,7 +255,6 @@ can not be used with the current version of glib-compiled-resources, due to
 
         return dep_files, depends, subdirs
 
-
     def _get_link_args(self, state, lib, depends=None, include_rpath=False):
         if gir_has_extra_lib_arg():
             link_command = ['--extra-library=%s' % lib.name]
@@ -258,42 +262,13 @@ can not be used with the current version of glib-compiled-resources, due to
             link_command = ['-l%s' % lib.name]
         if isinstance(lib, build.SharedLibrary):
             libdir = os.path.join(state.environment.get_build_dir(), lib.subdir)
-            link_command += ['-L%s' %libdir]
+            link_command += ['-L%s' % libdir]
             if include_rpath:
-                link_command += ['-Wl,-rpath %s' %libdir]
+                link_command += ['-Wl,-rpath %s' % libdir]
             if depends:
                 depends.append(lib)
         return link_command
 
-    @staticmethod
-    def _get_include_args(state, include_dirs, prefix='-I'):
-        if not include_dirs:
-            return []
-
-        dirs_str = []
-        for incdirs in include_dirs:
-            if hasattr(incdirs, "held_object"):
-                dirs = incdirs.held_object
-            else:
-                dirs = incdirs
-
-            if isinstance(dirs, str):
-                dirs_str += ['%s%s' % (prefix, dirs)]
-                continue
-
-            # Should be build.IncludeDirs object.
-            basedir = dirs.get_curdir()
-            for d in dirs.get_incdirs():
-                expdir =  os.path.join(basedir, d)
-                srctreedir = os.path.join(state.environment.get_source_dir(), expdir)
-                buildtreedir = os.path.join(state.environment.get_build_dir(), expdir)
-                dirs_str += ['%s%s' % (prefix, buildtreedir),
-                             '%s%s' % (prefix, srctreedir)]
-            for d in dirs.get_extra_build_dirs():
-                dirs_str += ['%s%s' % (prefix, d)]
-
-        return dirs_str
-
     def _get_dependencies_flags(self, deps, state, depends=None, include_rpath=False):
         cflags = set()
         ldflags = set()
@@ -305,7 +280,7 @@ can not be used with the current version of glib-compiled-resources, due to
             if hasattr(dep, 'held_object'):
                 dep = dep.held_object
             if isinstance(dep, InternalDependency):
-                cflags.update(self._get_include_args(state, dep.include_directories))
+                cflags.update(get_include_args(state.environment, dep.include_directories))
                 for lib in dep.libraries:
                     ldflags.update(self._get_link_args(state, lib.held_object, depends, include_rpath))
                     libdepflags = self._get_dependencies_flags(lib.held_object.get_external_deps(), state, depends, include_rpath)
@@ -319,7 +294,7 @@ can not be used with the current version of glib-compiled-resources, due to
                 for source in dep.sources:
                     if hasattr(source, 'held_object') and isinstance(source.held_object, GirTarget):
                         gi_includes.update([os.path.join(state.environment.get_build_dir(),
-                                        source.held_object.get_subdir())])
+                                            source.held_object.get_subdir())])
             # This should be any dependency other than an internal one.
             elif isinstance(dep, Dependency):
                 cflags.update(dep.get_compile_args())
@@ -379,7 +354,7 @@ can not be used with the current version of glib-compiled-resources, due to
             if not girwarning_printed:
                 mlog.warning('gobject-introspection dependency was not found, disabling gir generation.')
                 girwarning_printed = True
-            return []
+            return ModuleReturnValue(None, [])
         ns = kwargs.pop('namespace')
         nsversion = kwargs.pop('nsversion')
         libsources = kwargs.pop('sources')
@@ -389,14 +364,14 @@ can not be used with the current version of glib-compiled-resources, due to
 
         scan_command = giscanner.get_command() + ['@INPUT@']
         scan_command += pkgargs
-        scan_command += ['--no-libtool', '--namespace='+ns, '--nsversion=' + nsversion, '--warn-all',
+        scan_command += ['--no-libtool', '--namespace=' + ns, '--nsversion=' + nsversion, '--warn-all',
                          '--output', '@OUTPUT@']
 
         extra_args = mesonlib.stringlistify(kwargs.pop('extra_args', []))
         scan_command += extra_args
         scan_command += ['-I' + os.path.join(state.environment.get_source_dir(), state.subdir),
                          '-I' + os.path.join(state.environment.get_build_dir(), state.subdir)]
-        scan_command += self._get_include_args(state, girtarget.get_include_dirs())
+        scan_command += get_include_args(state.environment, girtarget.get_include_dirs())
 
         if 'link_with' in kwargs:
             link_with = kwargs.pop('link_with')
@@ -521,30 +496,29 @@ can not be used with the current version of glib-compiled-resources, due to
             if not isinstance(incd.held_object, (str, build.IncludeDirs)):
                 raise MesonException(
                     'Gir include dirs should be include_directories().')
-        scan_command += self._get_include_args(state, inc_dirs)
-        scan_command += self._get_include_args(state, gir_inc_dirs + inc_dirs,
-                                               prefix='--add-include-path=')
+        scan_command += get_include_args(state.environment, inc_dirs)
+        scan_command += get_include_args(state.environment, gir_inc_dirs + inc_dirs,
+                                         prefix='--add-include-path=')
 
         if isinstance(girtarget, build.Executable):
             scan_command += ['--program', girtarget]
         elif isinstance(girtarget, build.SharedLibrary):
             libname = girtarget.get_basename()
             scan_command += ['--library', libname]
-        scankwargs = {'output' : girfile,
-                      'input' : libsources,
-                      'command' : scan_command,
-                      'depends' : depends,
-                     }
+        scankwargs = {'output': girfile,
+                      'input': libsources,
+                      'command': scan_command,
+                      'depends': depends}
         if kwargs.get('install'):
             scankwargs['install'] = kwargs['install']
             scankwargs['install_dir'] = kwargs.get('install_dir_gir',
-                os.path.join(state.environment.get_datadir(), 'gir-1.0'))
+                                                   os.path.join(state.environment.get_datadir(), 'gir-1.0'))
         scan_target = GirTarget(girfile, state.subdir, scankwargs)
 
         typelib_output = '%s-%s.typelib' % (ns, nsversion)
         typelib_cmd = gicompiler.get_command() + [scan_target, '--output', '@OUTPUT@']
-        typelib_cmd += self._get_include_args(state, gir_inc_dirs,
-                                              prefix='--includedir=')
+        typelib_cmd += get_include_args(state.environment, gir_inc_dirs,
+                                        prefix='--includedir=')
         for incdir in typelib_includes:
             typelib_cmd += ["--includedir=" + incdir]
 
@@ -555,9 +529,10 @@ can not be used with the current version of glib-compiled-resources, due to
         if kwargs.get('install'):
             typelib_kwargs['install'] = kwargs['install']
             typelib_kwargs['install_dir'] = kwargs.get('install_dir_typelib',
-                os.path.join(state.environment.get_libdir(), 'girepository-1.0'))
+                                                       os.path.join(state.environment.get_libdir(), 'girepository-1.0'))
         typelib_target = TypelibTarget(typelib_output, state.subdir, typelib_kwargs)
-        return [scan_target, typelib_target]
+        rv = [scan_target, typelib_target]
+        return ModuleReturnValue(rv, rv)
 
     def compile_schemas(self, state, args, kwargs):
         if len(args) != 0:
@@ -575,7 +550,7 @@ can not be used with the current version of glib-compiled-resources, due to
         else:
             targetname = 'gsettings-compile-' + state.subdir
         target_g = build.CustomTarget(targetname, state.subdir, kwargs)
-        return target_g
+        return ModuleReturnValue(target_g, [target_g])
 
     def yelp(self, state, args, kwargs):
         if len(args) < 1:
@@ -602,13 +577,12 @@ can not be used with the current version of glib-compiled-resources, due to
 
         script = [sys.executable, state.environment.get_build_command()]
         args = ['--internal',
-            'yelphelper',
-            'install',
-            '--subdir=' + state.subdir,
-            '--id=' + project_id,
-            '--installdir=' + os.path.join(state.environment.get_datadir(), 'help'),
-            '--sources=' + source_str,
-        ]
+                'yelphelper',
+                'install',
+                '--subdir=' + state.subdir,
+                '--id=' + project_id,
+                '--installdir=' + os.path.join(state.environment.get_datadir(), 'help'),
+                '--sources=' + source_str]
         if symlinks:
             args.append('--symlinks=true')
         if media:
@@ -622,17 +596,18 @@ can not be used with the current version of glib-compiled-resources, due to
                    '--id=' + project_id,
                    '--sources=' + source_str]
         pottarget = build.RunTarget('help-' + project_id + '-pot', sys.executable,
-                                     potargs, [], state.subdir)
+                                    potargs, [], state.subdir)
 
         poargs = [state.environment.get_build_command(), '--internal', 'yelphelper', 'update-po',
-                   '--subdir=' + state.subdir,
-                   '--id=' + project_id,
-                   '--sources=' + source_str,
-                   '--langs=' + '@@'.join(langs)]
+                  '--subdir=' + state.subdir,
+                  '--id=' + project_id,
+                  '--sources=' + source_str,
+                  '--langs=' + '@@'.join(langs)]
         potarget = build.RunTarget('help-' + project_id + '-update-po', sys.executable,
-                                     poargs, [], state.subdir)
+                                   poargs, [], state.subdir)
 
-        return [inscript, pottarget, potarget]
+        rv = [inscript, pottarget, potarget]
+        return ModuleReturnValue(None, rv)
 
     def gtkdoc(self, state, args, kwargs):
         if len(args) != 1:
@@ -640,7 +615,7 @@ can not be used with the current version of glib-compiled-resources, due to
         modulename = args[0]
         if not isinstance(modulename, str):
             raise MesonException('Gtkdoc arg must be string.')
-        if not 'src_dir' in kwargs:
+        if 'src_dir' not in kwargs:
             raise MesonException('Keyword argument src_dir missing.')
         main_file = kwargs.get('main_sgml', '')
         if not isinstance(main_file, str):
@@ -658,7 +633,7 @@ can not be used with the current version of glib-compiled-resources, due to
         namespace = kwargs.get('namespace', '')
         mode = kwargs.get('mode', 'auto')
         VALID_MODES = ('xml', 'sgml', 'none', 'auto')
-        if not mode in VALID_MODES:
+        if mode not in VALID_MODES:
             raise MesonException('gtkdoc: Mode {} is not a valid mode: {}'.format(mode, VALID_MODES))
 
         src_dirs = kwargs['src_dir']
@@ -700,7 +675,7 @@ can not be used with the current version of glib-compiled-resources, due to
         res = [build.RunTarget(targetname, command[0], command[1:] + args, [], state.subdir)]
         if kwargs.get('install', True):
             res.append(build.RunScript(command, args))
-        return res
+        return ModuleReturnValue(None, res)
 
     def _get_build_args(self, kwargs, state):
         args = []
@@ -712,7 +687,7 @@ can not be used with the current version of glib-compiled-resources, due to
             if not isinstance(incd.held_object, (str, build.IncludeDirs)):
                 raise MesonException(
                     'Gir include dirs should be include_directories().')
-        cflags.update(self._get_include_args(state, inc_dirs))
+        cflags.update(get_include_args(state.environment, inc_dirs))
         if cflags:
             args += ['--cflags=%s' % ' '.join(cflags)]
         if ldflags:
@@ -730,7 +705,7 @@ can not be used with the current version of glib-compiled-resources, due to
         modulename = args[0]
         if not isinstance(modulename, str):
             raise MesonException('Argument must be a string')
-        return os.path.join('share/gtkdoc/html', modulename)
+        return ModuleReturnValue(os.path.join('share/gtkdoc/html', modulename), [])
 
     @staticmethod
     def _unpack_args(arg, kwarg_name, kwargs, expend_file_state=None):
@@ -766,11 +741,12 @@ can not be used with the current version of glib-compiled-resources, due to
             cmd += ['--c-namespace', kwargs.pop('namespace')]
         cmd += ['--generate-c-code', '@OUTDIR@/' + namebase, '@INPUT@']
         outputs = [namebase + '.c', namebase + '.h']
-        custom_kwargs = {'input' : xml_file,
-                         'output' : outputs,
-                         'command' : cmd
+        custom_kwargs = {'input': xml_file,
+                         'output': outputs,
+                         'command': cmd
                          }
-        return build.CustomTarget(target_name, state.subdir, custom_kwargs)
+        ct = build.CustomTarget(target_name, state.subdir, custom_kwargs)
+        return ModuleReturnValue(ct, [ct])
 
     def mkenums(self, state, args, kwargs):
         if len(args) != 1:
@@ -865,11 +841,11 @@ can not be used with the current version of glib-compiled-resources, due to
                     state.environment.coredata.get_builtin_option('includedir')
             target = self._make_mkenum_custom_target(state, sources, basename,
                                                      generic_cmd, custom_kwargs)
-            return target
+            return ModuleReturnValue(target, [target])
         elif len(targets) == 1:
-            return targets[0]
+            return ModuleReturnValue(targets[0], [targets[0]])
         else:
-            return targets
+            return ModuleReturnValue(targets, targets)
 
     @staticmethod
     def _make_mkenum_custom_target(state, sources, output, cmd, kwargs):
@@ -936,7 +912,8 @@ can not be used with the current version of glib-compiled-resources, due to
         custom_kwargs['output'] = output + '.h'
         header = build.CustomTarget(output + '_h', state.subdir, custom_kwargs)
 
-        return [body, header]
+        rv = [body, header]
+        return ModuleReturnValue(rv, rv)
 
     @staticmethod
     def _vapi_args_to_command(prefix, variable, kwargs, accept_vapi=False):
@@ -1020,6 +997,7 @@ can not be used with the current version of glib-compiled-resources, due to
 
         if not isinstance(args[0], str):
             raise MesonException('The first argument must be the name of the library')
+        created_values = []
 
         library = args[0]
         build_dir = os.path.join(state.environment.get_build_dir(), state.subdir)
@@ -1044,8 +1022,7 @@ can not be used with the current version of glib-compiled-resources, due to
         for i in inputs:
             if isinstance(i, str):
                 cmd.append(os.path.join(source_dir, i))
-            elif hasattr(i, 'held_object') \
-                 and isinstance(i.held_object, GirTarget):
+            elif hasattr(i, 'held_object') and isinstance(i.held_object, GirTarget):
                 link_with += self._get_vapi_link_with(i.held_object)
                 subdir = os.path.join(state.environment.get_build_dir(),
                                       i.held_object.get_subdir())
@@ -1062,16 +1039,15 @@ can not be used with the current version of glib-compiled-resources, due to
             'depends': vapi_depends,
         }
         install_dir = kwargs.get('install_dir',
-                os.path.join(state.environment.coredata.get_builtin_option('datadir'),
-                             'vala', 'vapi'))
+                                 os.path.join(state.environment.coredata.get_builtin_option('datadir'),
+                                              'vala', 'vapi'))
         if kwargs.get('install'):
             custom_kwargs['install'] = kwargs['install']
             custom_kwargs['install_dir'] = install_dir
 
             # We shouldn't need this locally but we install it
             deps_target = self._generate_deps(state, library, vapi_packages, install_dir)
-            # XXX WRONG, state objects must not be modified! Fix this!
-            state.data.append(deps_target)
+            created_values.append(deps_target)
         vapi_target = VapiTarget(vapi_output, state.subdir, custom_kwargs)
 
         # So to try our best to get this to just work we need:
@@ -1080,7 +1056,9 @@ can not be used with the current version of glib-compiled-resources, due to
         # - add relevant directories to include dirs
         incs = [build.IncludeDirs(state.subdir, ['.'] + vapi_includes, False)]
         sources = [vapi_target] + vapi_depends
-        return InternalDependency(None, incs, [], [], link_with, sources, [])
+        rv = InternalDependency(None, incs, [], [], link_with, sources, [])
+        created_values.append(rv)
+        return ModuleReturnValue(rv, created_values)
 
 def initialize():
     return GnomeModule()
