@@ -12,12 +12,16 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from .. import coredata, build
-from .. import mesonlib
-from .. import mlog
 import os
 
-class PkgConfigModule:
+from .. import build
+from .. import mesonlib
+from .. import mlog
+from . import ModuleReturnValue
+from . import ExtensionModule
+
+
+class PkgConfigModule(ExtensionModule):
 
     def _get_lname(self, l, msg, pcfile):
         # Nothing special
@@ -48,7 +52,7 @@ class PkgConfigModule:
             # '${prefix}' is ignored if the second path is absolute (see
             # 'os.path.join' for details)
             ofile.write('libdir=%s\n' % os.path.join('${prefix}', coredata.get_builtin_option('libdir')))
-            ofile.write('includedir=%s\n' %  os.path.join('${prefix}', coredata.get_builtin_option('includedir')))
+            ofile.write('includedir=%s\n' % os.path.join('${prefix}', coredata.get_builtin_option('includedir')))
             ofile.write('\n')
             ofile.write('Name: %s\n' % name)
             if len(description) > 0:
@@ -63,6 +67,7 @@ class PkgConfigModule:
                     'Requires.private: {}\n'.format(' '.join(priv_reqs)))
             if len(conflicts) > 0:
                 ofile.write('Conflicts: {}\n'.format(' '.join(conflicts)))
+
             def generate_libs_flags(libs):
                 msg = 'Library target {0!r} has {1!r} set. Compilers ' \
                       'may not find it from its \'-l{2}\' linker flag in the ' \
@@ -81,6 +86,7 @@ class PkgConfigModule:
                         if l.name_suffix_set:
                             mlog.warning(msg.format(l.name, 'name_suffix', lname, pcfile))
                         yield '-l%s' % lname
+
             if len(libraries) > 0:
                 ofile.write('Libs: {}\n'.format(' '.join(generate_libs_flags(libraries))))
             if len(priv_libs) > 0:
@@ -130,7 +136,7 @@ class PkgConfigModule:
         priv_reqs = mesonlib.stringlistify(kwargs.get('requires_private', []))
         conflicts = mesonlib.stringlistify(kwargs.get('conflicts', []))
         pcfile = filebase + '.pc'
-        pkgroot = kwargs.get('install_dir',None)
+        pkgroot = kwargs.get('install_dir', None)
         if pkgroot is None:
             pkgroot = os.path.join(state.environment.coredata.get_builtin_option('libdir'), 'pkgconfig')
         if not isinstance(pkgroot, str):
@@ -138,7 +144,8 @@ class PkgConfigModule:
         self.generate_pkgconfig_file(state, libs, subdirs, name, description, url,
                                      version, pcfile, pub_reqs, priv_reqs,
                                      conflicts, priv_libs)
-        return build.Data(mesonlib.File(True, state.environment.get_scratch_dir(), pcfile), pkgroot)
+        res = build.Data(mesonlib.File(True, state.environment.get_scratch_dir(), pcfile), pkgroot)
+        return ModuleReturnValue(res, [res])
 
 def initialize():
     return PkgConfigModule()
