@@ -12,14 +12,16 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import os, subprocess
+import os
 from .. import mlog
 from .. import build
-from ..mesonlib import MesonException
+from ..mesonlib import MesonException, Popen_safe
 from ..dependencies import Qt5Dependency
+from . import ExtensionModule
 import xml.etree.ElementTree as ET
+from . import ModuleReturnValue
 
-class Qt5Module():
+class Qt5Module(ExtensionModule):
     tools_detected = False
 
     def _detect_tools(self, env):
@@ -37,53 +39,47 @@ class Qt5Module():
         # Moc and rcc return a non-zero result when doing so.
         # What kind of an idiot thought that was a good idea?
         if self.moc.found():
-            mp = subprocess.Popen(self.moc.get_command() + ['-v'],
-                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-            (stdout, stderr) = mp.communicate()
-            stdout = stdout.decode().strip()
-            stderr = stderr.decode().strip()
+            stdout, stderr = Popen_safe(self.moc.get_command() + ['-v'])[1:3]
+            stdout = stdout.strip()
+            stderr = stderr.strip()
             if 'Qt 5' in stderr:
                 moc_ver = stderr
             elif '5.' in stdout:
                 moc_ver = stdout
             else:
                 raise MesonException('Moc preprocessor is not for Qt 5. Output:\n%s\n%s' %
-                                          (stdout, stderr))
-            mlog.log(' moc:', mlog.green('YES'), '(%s, %s)' % \
+                                     (stdout, stderr))
+            mlog.log(' moc:', mlog.green('YES'), '(%s, %s)' %
                      (' '.join(self.moc.fullpath), moc_ver.split()[-1]))
         else:
             mlog.log(' moc:', mlog.red('NO'))
         if self.uic.found():
-            up = subprocess.Popen(self.uic.get_command() + ['-v'],
-                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-            (stdout, stderr) = up.communicate()
-            stdout = stdout.decode().strip()
-            stderr = stderr.decode().strip()
+            stdout, stderr = Popen_safe(self.uic.get_command() + ['-v'])[1:3]
+            stdout = stdout.strip()
+            stderr = stderr.strip()
             if 'version 5.' in stderr:
                 uic_ver = stderr
             elif '5.' in stdout:
                 uic_ver = stdout
             else:
                 raise MesonException('Uic compiler is not for Qt 5. Output:\n%s\n%s' %
-                                          (stdout, stderr))
-            mlog.log(' uic:', mlog.green('YES'), '(%s, %s)' % \
+                                     (stdout, stderr))
+            mlog.log(' uic:', mlog.green('YES'), '(%s, %s)' %
                      (' '.join(self.uic.fullpath), uic_ver.split()[-1]))
         else:
             mlog.log(' uic:', mlog.red('NO'))
         if self.rcc.found():
-            rp = subprocess.Popen(self.rcc.get_command() + ['-v'],
-                                  stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-            (stdout, stderr) = rp.communicate()
-            stdout = stdout.decode().strip()
-            stderr = stderr.decode().strip()
+            stdout, stderr = Popen_safe(self.rcc.get_command() + ['-v'])[1:3]
+            stdout = stdout.strip()
+            stderr = stderr.strip()
             if 'version 5.' in stderr:
                 rcc_ver = stderr
             elif '5.' in stdout:
                 rcc_ver = stdout
             else:
                 raise MesonException('Rcc compiler is not for Qt 5. Output:\n%s\n%s' %
-                                          (stdout, stderr))
-            mlog.log(' rcc:', mlog.green('YES'), '(%s, %s)'\
+                                     (stdout, stderr))
+            mlog.log(' rcc:', mlog.green('YES'), '(%s, %s)'
                      % (' '.join(self.rcc.fullpath), rcc_ver.split()[-1]))
         else:
             mlog.log(' rcc:', mlog.red('NO'))
@@ -119,10 +115,10 @@ class Qt5Module():
         moc_sources = kwargs.pop('moc_sources', [])
         if not isinstance(moc_sources, list):
             moc_sources = [moc_sources]
-        srctmp = kwargs.pop('sources', [])
-        if not isinstance(srctmp, list):
-            srctmp = [srctmp]
-        sources = args[1:] + srctmp
+        sources = kwargs.pop('sources', [])
+        if not isinstance(sources, list):
+            sources = [sources]
+        sources += args[1:]
         self._detect_tools(state.environment)
         err_msg = "{0} sources specified and couldn't find {1}, " \
                   "please check your qt5 installation"
@@ -134,39 +130,38 @@ class Qt5Module():
             qrc_deps = []
             for i in rcc_files:
                 qrc_deps += self.parse_qrc(state, i)
-            basename = os.path.split(rcc_files[0])[1]
-            rcc_kwargs = {'input' : rcc_files,
-                    'output' : basename + '.cpp',
-                    'command' : [self.rcc, '-o', '@OUTPUT@', '@INPUT@'],
-                    'depend_files' : qrc_deps,
-                    }
-            name = 'qt5-' + basename.replace('.', '_')
+            if len(args) > 0:
+                name = args[0]
+            else:
+                basename = os.path.split(rcc_files[0])[1]
+                name = 'qt5-' + basename.replace('.', '_')
+            rcc_kwargs = {'input': rcc_files,
+                          'output': name + '.cpp',
+                          'command': [self.rcc, '-o', '@OUTPUT@', '@INPUT@'],
+                          'depend_files': qrc_deps}
             res_target = build.CustomTarget(name, state.subdir, rcc_kwargs)
             sources.append(res_target)
         if len(ui_files) > 0:
             if not self.uic.found():
                 raise MesonException(err_msg.format('UIC', 'uic-qt5'))
-            ui_kwargs = {'output' : 'ui_@BASENAME@.h',
-                         'arguments' : ['-o', '@OUTPUT@', '@INPUT@']}
+            ui_kwargs = {'output': 'ui_@BASENAME@.h',
+                         'arguments': ['-o', '@OUTPUT@', '@INPUT@']}
             ui_gen = build.Generator([self.uic], ui_kwargs)
-            ui_output = build.GeneratedList(ui_gen)
-            [ui_output.add_file(os.path.join(state.subdir, a)) for a in ui_files]
+            ui_output = ui_gen.process_files('Qt5 ui', ui_files, state)
             sources.append(ui_output)
         if len(moc_headers) > 0:
-            moc_kwargs = {'output' : 'moc_@BASENAME@.cpp',
-                          'arguments' : ['@INPUT@', '-o', '@OUTPUT@']}
+            moc_kwargs = {'output': 'moc_@BASENAME@.cpp',
+                          'arguments': ['@INPUT@', '-o', '@OUTPUT@']}
             moc_gen = build.Generator([self.moc], moc_kwargs)
-            moc_output = build.GeneratedList(moc_gen)
-            [moc_output.add_file(os.path.join(state.subdir, a)) for a in moc_headers]
+            moc_output = moc_gen.process_files('Qt5 moc header', moc_headers, state)
             sources.append(moc_output)
         if len(moc_sources) > 0:
-            moc_kwargs = {'output' : '@BASENAME@.moc',
-                          'arguments' : ['@INPUT@', '-o', '@OUTPUT@']}
+            moc_kwargs = {'output': '@BASENAME@.moc',
+                          'arguments': ['@INPUT@', '-o', '@OUTPUT@']}
             moc_gen = build.Generator([self.moc], moc_kwargs)
-            moc_output = build.GeneratedList(moc_gen)
-            [moc_output.add_file(os.path.join(state.subdir, a)) for a in moc_sources]
+            moc_output = moc_gen.process_files('Qt5 moc source', moc_sources, state)
             sources.append(moc_output)
-        return sources
+        return ModuleReturnValue(sources, sources)
 
 def initialize():
     mlog.warning('rcc dependencies will not work reliably until this upstream issue is fixed:',
