@@ -62,18 +62,28 @@ def ComboParser(name, description, kwargs):
             raise OptionException('Combo choice elements must be strings.')
     return coredata.UserComboOption(name, description, choices, kwargs.get('value', choices[0]))
 
-option_types = {'string' : StringParser,
-                'boolean' : BooleanParser,
-                'combo' : ComboParser,
-               }
+option_types = {'string': StringParser,
+                'boolean': BooleanParser,
+                'combo': ComboParser,
+                }
 
 class OptionInterpreter:
     def __init__(self, subproject, command_line_options):
         self.options = {}
         self.subproject = subproject
+        self.sbprefix = subproject + ':'
         self.cmd_line_options = {}
         for o in command_line_options:
-            (key, value) = o.split('=', 1)
+            if self.subproject != '': # Strip the beginning.
+                if not o.startswith(self.sbprefix):
+                    continue
+            else:
+                if ':' in o:
+                    continue
+            try:
+                (key, value) = o.split('=', 1)
+            except ValueError:
+                raise OptionException('Option {!r} must have a value separated by equals sign.'.format(o))
             self.cmd_line_options[key] = value
 
     def process(self, option_file):
@@ -118,7 +128,7 @@ class OptionInterpreter:
                 raise OptionException('Keyword argument name is not a string.')
             a = args.kwargs[key]
             reduced_kw[key] = self.reduce_single(a)
-        return (reduced_pos, reduced_kw)
+        return reduced_pos, reduced_kw
 
     def evaluate_statement(self, node):
         if not isinstance(node, mparser.FunctionNode):
@@ -130,7 +140,7 @@ class OptionInterpreter:
         if 'type' not in kwargs:
             raise OptionException('Option call missing mandatory "type" keyword argument')
         opt_type = kwargs['type']
-        if not opt_type in option_types:
+        if opt_type not in option_types:
             raise OptionException('Unknown type %s.' % opt_type)
         if len(posargs) != 1:
             raise OptionException('Option() must have one (and only one) positional argument')
