@@ -23,31 +23,31 @@ DT_STRTAB = 5
 DT_SONAME = 14
 DT_MIPS_RLD_MAP_REL = 1879048245
 
-class DataSizes():
+class DataSizes:
     def __init__(self, ptrsize, is_le):
         if is_le:
             p = '<'
         else:
             p = '>'
-        self.Half = p+'h'
+        self.Half = p + 'h'
         self.HalfSize = 2
-        self.Word = p+'I'
+        self.Word = p + 'I'
         self.WordSize = 4
-        self.Sword = p+'i'
+        self.Sword = p + 'i'
         self.SwordSize = 4
         if ptrsize == 64:
-            self.Addr = p+'Q'
+            self.Addr = p + 'Q'
             self.AddrSize = 8
-            self.Off = p+'Q'
+            self.Off = p + 'Q'
             self.OffSize = 8
-            self.XWord = p+'Q'
+            self.XWord = p + 'Q'
             self.XWordSize = 8
-            self.Sxword = p+'q'
+            self.Sxword = p + 'q'
             self.SxwordSize = 8
         else:
-            self.Addr = p+'I'
+            self.Addr = p + 'I'
             self.AddrSize = 4
-            self.Off = p+'I'
+            self.Off = p + 'I'
             self.OffSize = 4
 
 class DynamicEntry(DataSizes):
@@ -55,8 +55,8 @@ class DynamicEntry(DataSizes):
         super().__init__(ptrsize, is_le)
         self.ptrsize = ptrsize
         if ptrsize == 64:
-            self.d_tag = struct.unpack(self.Sxword, ifile.read(self.SxwordSize))[0];
-            self.val = struct.unpack(self.XWord, ifile.read(self.XWordSize))[0];
+            self.d_tag = struct.unpack(self.Sxword, ifile.read(self.SxwordSize))[0]
+            self.val = struct.unpack(self.XWord, ifile.read(self.XWordSize))[0]
         else:
             self.d_tag = struct.unpack(self.Sword, ifile.read(self.SwordSize))[0]
             self.val = struct.unpack(self.Word, ifile.read(self.WordSize))[0]
@@ -76,29 +76,29 @@ class SectionHeader(DataSizes):
             is_64 = True
         else:
             is_64 = False
-#Elf64_Word
-        self.sh_name = struct.unpack(self.Word, ifile.read(self.WordSize))[0];
-#Elf64_Word
+# Elf64_Word
+        self.sh_name = struct.unpack(self.Word, ifile.read(self.WordSize))[0]
+# Elf64_Word
         self.sh_type = struct.unpack(self.Word, ifile.read(self.WordSize))[0]
-#Elf64_Xword
+# Elf64_Xword
         if is_64:
             self.sh_flags = struct.unpack(self.XWord, ifile.read(self.XWordSize))[0]
         else:
             self.sh_flags = struct.unpack(self.Word, ifile.read(self.WordSize))[0]
-#Elf64_Addr
-        self.sh_addr = struct.unpack(self.Addr, ifile.read(self.AddrSize))[0];
-#Elf64_Off
+# Elf64_Addr
+        self.sh_addr = struct.unpack(self.Addr, ifile.read(self.AddrSize))[0]
+# Elf64_Off
         self.sh_offset = struct.unpack(self.Off, ifile.read(self.OffSize))[0]
-#Elf64_Xword
+# Elf64_Xword
         if is_64:
             self.sh_size = struct.unpack(self.XWord, ifile.read(self.XWordSize))[0]
         else:
             self.sh_size = struct.unpack(self.Word, ifile.read(self.WordSize))[0]
-#Elf64_Word
-        self.sh_link = struct.unpack(self.Word, ifile.read(self.WordSize))[0];
-#Elf64_Word
-        self.sh_info = struct.unpack(self.Word, ifile.read(self.WordSize))[0];
-#Elf64_Xword
+# Elf64_Word
+        self.sh_link = struct.unpack(self.Word, ifile.read(self.WordSize))[0]
+# Elf64_Word
+        self.sh_info = struct.unpack(self.Word, ifile.read(self.WordSize))[0]
+# Elf64_Xword
         if is_64:
             self.sh_addralign = struct.unpack(self.XWord, ifile.read(self.XWordSize))[0]
         else:
@@ -150,7 +150,7 @@ class Elf(DataSizes):
             is_le = False
         else:
             sys.exit('File "%s" has unknown ELF endianness.' % self.bfile)
-        return (ptrsize, is_le)
+        return ptrsize, is_le
 
     def parse_header(self):
         self.bf.seek(0)
@@ -196,6 +196,8 @@ class Elf(DataSizes):
     def parse_dynamic(self):
         sec = self.find_section(b'.dynamic')
         self.dynamic = []
+        if sec is None:
+            return
         self.bf.seek(sec.sh_offset)
         while True:
             e = DynamicEntry(self.bf, self.ptrsize, self.is_le)
@@ -218,6 +220,9 @@ class Elf(DataSizes):
                 soname = i
             if i.d_tag == DT_STRTAB:
                 strtab = i
+        if soname is None or strtab is None:
+            print("This file does not have a soname")
+            return
         self.bf.seek(strtab.val + soname.val)
         print(self.read_str())
 
@@ -268,7 +273,7 @@ class Elf(DataSizes):
             name = self.read_str()
             if name.startswith(prefix):
                 basename = name.split(b'/')[-1]
-                padding = b'\0'*(len(name) - len(basename))
+                padding = b'\0' * (len(name) - len(basename))
                 newname = basename + padding
                 assert(len(newname) == len(name))
                 self.bf.seek(offset)
@@ -294,18 +299,20 @@ class Elf(DataSizes):
             sys.exit("New rpath must not be longer than the old one.")
         self.bf.seek(rp_off)
         self.bf.write(new_rpath)
-        self.bf.write(b'\0'*(len(old_rpath) - len(new_rpath) + 1))
+        self.bf.write(b'\0' * (len(old_rpath) - len(new_rpath) + 1))
         if len(new_rpath) == 0:
             self.remove_rpath_entry(entrynum)
 
     def remove_rpath_entry(self, entrynum):
         sec = self.find_section(b'.dynamic')
+        if sec is None:
+            return None
         for (i, entry) in enumerate(self.dynamic):
             if entry.d_tag == entrynum:
                 rpentry = self.dynamic[i]
                 rpentry.d_tag = 0
-                self.dynamic = self.dynamic[:i] + self.dynamic[i+1:] + [rpentry]
-                break;
+                self.dynamic = self.dynamic[:i] + self.dynamic[i + 1:] + [rpentry]
+                break
         # DT_MIPS_RLD_MAP_REL is relative to the offset of the tag. Adjust it consequently.
         for entry in self.dynamic[i:]:
             if entry.d_tag == DT_MIPS_RLD_MAP_REL:
